from abc import ABCMeta, abstractmethod
import enum
from pathlib import Path
import tempfile
from urllib.parse import quote, unquote

from cumulusci.core.exceptions import CumulusCIException, TaskOptionsError
from cumulusci.core.tasks import BaseSalesforceTask
from cumulusci.salesforce_api.metadata import ApiRetrieveUnpackaged
from cumulusci.tasks.metadata.package import PackageXmlGenerator
from cumulusci.core.utils import process_bool_arg, process_list_arg
from cumulusci.utils import inject_namespace
from cumulusci.core.config import TaskConfig
from cumulusci.utils.xml import metadata_tree
from cumulusci.utils.xml.metadata_tree import MetadataElement


class MetadataOperation(enum.Enum):
    DEPLOY = "deploy"
    RETRIEVE = "retrieve"


class BaseMetadataETLTask(BaseSalesforceTask, metaclass=ABCMeta):
    """Abstract base class for all Metadata ETL tasks. Concrete tasks should
    generally subclass BaseMetadataSynthesisTask, BaseMetadataTransformTask,
    or MetadataSingleEntityTransformTask."""

    deploy = False
    retrieve = False

    task_options = {
        "managed": {
            "description": "If False, changes namespace_inject to replace tokens with a blank string"
        },
        "namespace_inject": {
            "description": "If set, the namespace tokens in files and filenames are replaced with the namespace's prefix"
        },
        "api_version": {
            "description": "Metadata API version to use, if not project__package__api_version."
        },
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)

        self.options["managed"] = process_bool_arg(self.options.get("managed", False))
        self.api_version = (
            self.options.get("api_version")
            or self.project_config.project__package__api_version
        )
        try:
            float(self.api_version)
        except ValueError:
            raise TaskOptionsError(f"Invalid API version {self.api_version}")

    def _inject_namespace(self, text):
        """Inject the namespace into the given text if running in managed mode."""
        return inject_namespace(
            "", text, self.options.get("namespace_inject"), self.options["managed"]
        )[1]

    @abstractmethod
    def _get_package_xml_content(self, operation):
        """Return the textual content of a package.xml for the given operation."""
        pass

    def _generate_package_xml(self, operation):
        """Call _get_package_xml_content() and perform namespace injection if needed"""
        return self._inject_namespace(self._get_package_xml_content(operation))

    def _create_directories(self, tempdir):
        """Create self.retrieve_dir and self.deploy_dir, if required"""
        if self.retrieve:
            self.retrieve_dir = Path(tempdir, "retrieve")
            self.retrieve_dir.mkdir()
        if self.deploy:
            self.deploy_dir = Path(tempdir, "deploy")
            self.deploy_dir.mkdir()

    def _retrieve(self):
        """Retrieve metadata into self.retrieve_dir"""
        self.logger.info("Extracting existing metadata...")
        api_retrieve = ApiRetrieveUnpackaged(
            self,
            self._generate_package_xml(MetadataOperation.RETRIEVE),
            self.api_version,
        )
        unpackaged = api_retrieve()
        unpackaged.extractall(self.retrieve_dir)

    @abstractmethod
    def _transform(self):
        """Transform the metadata in self.retrieve_dir into self.deploy_dir."""
        pass

    def _deploy(self):
        """Deploy metadata from self.deploy_dir"""
        self.logger.info("Loading transformed metadata...")
        target_profile_xml = Path(self.deploy_dir, "package.xml")
        target_profile_xml.write_text(
            self._generate_package_xml(MetadataOperation.DEPLOY), encoding="utf-8"
        )

        # import is here to avoid an import cycle
        from cumulusci.tasks.salesforce import Deploy

        api = Deploy(
            self.project_config,
            TaskConfig(
                {
                    "options": {
                        "path": self.deploy_dir,
                        "namespace_inject": self.options.get("namespace_inject"),
                        "unmanaged": not self.options.get("managed"),
                    }
                }
            ),
            self.org_config,
        )
        result = api()

        return result

    def _post_deploy(self, result):
        """Run any post-deploy logic required, such as waiting for asynchronous
        operations to complete in the target org."""
        pass

    def _run_task(self):
        with tempfile.TemporaryDirectory() as tempdir:
            self._create_directories(tempdir)
            if self.retrieve:
                self._retrieve()
            self._transform()
            if self.deploy:
                result = self._deploy()
                self._post_deploy(result)


class BaseMetadataSynthesisTask(BaseMetadataETLTask, metaclass=ABCMeta):
    """Base class for Metadata ETL tasks that generate new metadata
    and deploy it into the org, but do not retrieve."""

    deploy = True

    def _generate_package_xml(self, deploy):
        """Synthesize a package.xml for generated metadata."""
        generator = PackageXmlGenerator(str(self.deploy_dir), self.api_version)
        return generator()

    def _transform(self):
        self._synthesize()

    @abstractmethod
    def _synthesize(self):
        """Create new metadata in self.deploy_dir."""
        pass


class BaseMetadataTransformTask(BaseMetadataETLTask, metaclass=ABCMeta):
    """Base class for Metadata ETL tasks that extract metadata,
    transform it, and deploy it back into the org."""

    retrieve = True
    deploy = True

    @abstractmethod
    def _get_entities(self):
        """Return a dict of Metadata API entities and API names to be transformed."""
        pass

    def _get_types_package_xml(self):
        """Generate package.xml content based on the return value of _get_entities()."""
        base = """    <types>
{members}
        <name>{name}</name>
    </types>
"""
        types = ""
        for entity, api_names in self._get_entities().items():
            members = "\n".join(
                f"        <members>{api_name}</members>"
                for api_name in sorted(api_names)
            )
            types += base.format(members=members, name=entity)

        return types

    def _get_package_xml_content(self, operation):
        return f"""<?xml version="1.0" encoding="UTF-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
{self._get_types_package_xml()}
    <version>{self.api_version}</version>
</Package>
"""

    @abstractmethod
    def _transform(self):
        pass


class MetadataSingleEntityTransformTask(BaseMetadataTransformTask, metaclass=ABCMeta):
    """Base class for a Metadata ETL task that affects one or more
    instances of a specific metadata entity. Concrete subclasses must set
    `entity` to the Metadata API entity transformed, and implement _transform_entity()."""

    entity = None

    task_options = {
        "api_names": {"description": "List of API names of entities to affect"},
        **BaseMetadataETLTask.task_options,
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)

        self.api_names = {
            self._inject_namespace(arg)
            for arg in process_list_arg(self.options.get("api_names", ["*"]))
        }
        self.api_names = {
            quote(arg, safe=" ") if arg != "*" else arg for arg in self.api_names
        }

    def _get_entities(self):
        return {self.entity: self.api_names}

    @abstractmethod
    def _transform_entity(self, metadata, api_name):
        """Accept an XML element corresponding to the metadata entity with
        the given api_name. Transform the XML and return the version which
        should be deployed, or None to suppress deployment of this entity."""
        pass

    def _transform(self):
        # call _transform_entity once per retrieved entity
        # if the entity is an XML file, provide a parsed version
        # and write the returned metadata into the deploy directory

        parser = PackageXmlGenerator(
            None, self.api_version
        )  # We'll use it for its metadata_map
        entity_configurations = [
            entry
            for entry in parser.metadata_map
            if any(
                [
                    subentry["type"] == self.entity
                    for subentry in parser.metadata_map[entry]
                ]
            )
        ]
        if not entity_configurations:
            raise CumulusCIException(
                f"Unable to locate configuration for entity {self.entity}"
            )

        configuration = parser.metadata_map[entity_configurations[0]][0]
        if configuration["class"] not in [
            "MetadataFilenameParser",
            "CustomObjectParser",
        ]:
            raise CumulusCIException(
                f"MetadataSingleEntityTransformTask only supports manipulating complete, file-based XML entities (not {self.entity})"
            )

        extension = configuration["extension"]
        directory = entity_configurations[0]
        source_metadata_dir = self.retrieve_dir / directory

        if "*" in self.api_names:
            # Walk the retrieved directory to get the actual suite
            # of API names retrieved and rebuild our api_names list.
            self.api_names.remove("*")
            self.api_names = self.api_names.union(
                metadata_file.stem
                for metadata_file in source_metadata_dir.iterdir()
                if metadata_file.suffix == f".{extension}"
            )

        removed_api_names = set()

        for api_name in self.api_names:
            # Page Layout names can contain spaces, but parentheses and other
            # characters like ' and < are quoted.
            # We quote user-specified API names so we can locate the corresponding
            # metadata files, but present them un-quoted in messages to the user.
            unquoted_api_name = unquote(api_name)

            path = source_metadata_dir / f"{api_name}.{extension}"
            if not path.exists():
                raise CumulusCIException(f"Cannot find metadata file {path}")

            try:
                tree = metadata_tree.parse(str(path))
            except SyntaxError as err:
                err.filename = path
                raise err
            transformed_xml = self._transform_entity(tree, unquoted_api_name)
            if transformed_xml:
                parent_dir = self.deploy_dir / directory
                if not parent_dir.exists():
                    parent_dir.mkdir()
                destination_path = parent_dir / f"{api_name}.{extension}"

                with destination_path.open(mode="w", encoding="utf-8") as f:
                    f.write(transformed_xml.tostring(xml_declaration=True))
            else:
                # Make sure to remove from our package.xml
                removed_api_names.add(api_name)

        self.api_names = self.api_names - removed_api_names


class UpdateMetadataFirstChildTextTask(MetadataSingleEntityTransformTask):
    task_docs = """
Metadata ETL task to update a single child element's text within metadata XML.

If the child doesn't exist, the child is created and appended to the Metadata.   Furthermore, the ``value`` option is namespaced injected if the task is properly configured.

Example: Assign a Custom Object's Compact Layout
------------------------------------------------

Researching `CustomObject <https://developer.salesforce.com/docs/atlas.en-us.api_meta.meta/api_meta/customobject.htm>`_ in the Metadata API documentation or even retrieving the CustomObject's Metadata for inspection, we see the ``compactLayoutAssignment`` Field.  We want to assign a specific Compact Layout for our Custom Object, so we write the following CumulusCI task in our project's ``cumulusci.yml``.

.. code-block::  yaml

  tasks:
      assign_compact_layout:
          class_path: cumulusci.tasks.metadata_etl.UpdateMetadataFirstChildTextTask
          options:
              managed: False
              namespace_inject: $project_config.project__package__namespace
              entity: CustomObject
              api_names: OurCustomObject__c
              tag: compactLayoutAssignment
              value: "%%%NAMESPACE%%%DifferentCompactLayout"
              # We include a namespace token so it's easy to use this task in a managed context.

Suppose the original CustomObject metadata XML looks like:

.. code-block:: xml

  <?xml version="1.0" encoding="UTF-8"?>
  <CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
      ...
      <label>Our Custom Object</label>
      <compactLayoutAssignment>OriginalCompactLayout</compactLayoutAssignment>
      ...
  </CustomObject>

After running ``cci task run assign_compact_layout``, the CustomObject metadata XML is deployed as:

.. code-block:: xml

  <?xml version="1.0" encoding="UTF-8"?>
  <CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
      ...
      <label>Our Custom Object</label>
      <compactLayoutAssignment>DifferentCompactLayout</compactLayoutAssignment>
      ...
  </CustomObject>
    """

    task_options = {
        "metadata_type": {"description": "Metadata Type", "required": True},
        "tag": {
            "description": "Targeted tag. The text of the first instance of this tag within the metadata entity will be updated.",
            "required": True,
        },
        "value": {
            "description": "Desired value to set for the targeted tag's text. This value is namespace-injected.",
            "required": True,
        },
        **MetadataSingleEntityTransformTask.task_options,
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        self.entity = self.options.get("metadata_type")
        self.options["value"] = self._inject_namespace(self.options.get("value"))

    def _transform_entity(
        self, metadata: MetadataElement, api_name: str
    ) -> MetadataElement:
        """Finds metadata's first child with tag.  If no child is found, appends
        a new child with tag.  Then updates child's text as the value option."""
        tag = self.options["tag"]

        child = metadata.find(tag)
        if child is None:
            child = metadata.append(tag)

        child.text = self.options["value"]

        self.logger.info(f'Updating {self.entity} "{api_name}":')
        self.logger.info(f'    {tag} as "{child.text}"')

        return metadata



from cumulusci.core.exceptions import DependencyLookupError
from datetime import datetime

import pytest
import responses

from cumulusci.core.config import OrgConfig
from cumulusci.core.config import ServiceConfig
from cumulusci.core.config import TaskConfig
from cumulusci.tests.util import create_project_config
from cumulusci.tasks.github.commit_status import GetPackageDataFromCommitStatus
from cumulusci.tasks.github.tests.util_github_api import GithubApiTestMixin


class TestGetPackageDataFromCommitStatus(GithubApiTestMixin):
    @responses.activate
    def test_run_task(self):
        self.init_github()
        repo_response = self._get_expected_repo("TestOwner", "TestRepo")
        now = datetime.now().isoformat()
        responses.add(method=responses.GET, url=self.repo_api_url, json=repo_response)
        responses.add(
            method=responses.GET,
            url=f"{self.repo_api_url}/commits/abcdef",
            json=self._get_expected_commit("abcdef"),
        )
        responses.add(
            method=responses.GET,
            url=f"{self.repo_api_url}/commits/abcdef/status",
            json={
                "url": f"{self.repo_api_url}/abcdef/status",
                "commit_url": f"{self.repo_api_url}/abcdef",
                "repository": repo_response,
                "sha": "abcdef",
                "state": "success",
                "statuses": [
                    {
                        "context": "2gp",
                        "created_at": now,
                        "updated_at": now,
                        "description": "version_id: 04t_1",
                        "id": 1,
                        "state": "success",
                        "target_url": None,
                        "url": f"{self.repo_api_url}/statuses/abcdef",
                    }
                ],
                "total_count": 1,
            },
        )
        responses.add(
            "GET",
            "https://salesforce/services/data/v49.0/tooling/query/",
            json={
                "records": [
                    {"Dependencies": {"ids": [{"subscriberPackageVersionId": "04t_2"}]}}
                ]
            },
        )

        project_config = create_project_config(repo_commit="abcdef")
        project_config.keychain.set_service(
            "github",
            ServiceConfig(
                {
                    "username": "TestUser",
                    "password": "TestPass",
                    "email": "testuser@testdomain.com",
                }
            ),
        )
        task_config = TaskConfig({"options": {"context": "2gp"}})
        org_config = OrgConfig(
            {"instance_url": "https://salesforce", "access_token": "TOKEN"}, "test"
        )
        task = GetPackageDataFromCommitStatus(project_config, task_config, org_config)
        task._init_task()
        task._run_task()
        assert task.return_values == {
            "version_id": "04t_1",
            "dependencies": [{"version_id": "04t_2"}],
        }

    @responses.activate
    def test_run_task__commit_not_found(self):
        self.init_github()
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo("TestOwner", "TestRepo"),
        )
        responses.add(
            method=responses.GET, url=f"{self.repo_api_url}/commits/abcdef", status=404
        )

        project_config = create_project_config(repo_commit="abcdef")
        project_config.keychain.set_service(
            "github",
            ServiceConfig(
                {
                    "username": "TestUser",
                    "password": "TestPass",
                    "email": "testuser@testdomain.com",
                }
            ),
        )
        task_config = TaskConfig({"options": {"context": "2gp"}})
        org_config = OrgConfig(
            {"instance_url": "https://salesforce", "access_token": "TOKEN"}, "test"
        )
        task = GetPackageDataFromCommitStatus(project_config, task_config, org_config)
        task._init_task()
        with pytest.raises(
            DependencyLookupError, match="Could not find commit abcdef on github"
        ):
            task._run_task()

    @responses.activate
    def test_run_task__status_not_found(self):
        self.init_github()
        repo_response = self._get_expected_repo("TestOwner", "TestRepo")
        responses.add(method=responses.GET, url=self.repo_api_url, json=repo_response)
        responses.add(
            method=responses.GET,
            url=f"{self.repo_api_url}/commits/abcdef",
            json=self._get_expected_commit("abcdef"),
        )
        responses.add(
            method=responses.GET,
            url=f"{self.repo_api_url}/commits/abcdef/status",
            json={
                "url": f"{self.repo_api_url}/abcdef/status",
                "commit_url": f"{self.repo_api_url}/abcdef",
                "repository": repo_response,
                "sha": "abcdef",
                "state": "success",
                "statuses": [],
                "total_count": 0,
            },
        )

        project_config = create_project_config(repo_commit="abcdef")
        project_config.keychain.set_service(
            "github",
            ServiceConfig(
                {
                    "username": "TestUser",
                    "password": "TestPass",
                    "email": "testuser@testdomain.com",
                }
            ),
        )
        task_config = TaskConfig({"options": {"context": "2gp"}})
        org_config = OrgConfig(
            {"instance_url": "https://salesforce", "access_token": "TOKEN"}, "test"
        )
        task = GetPackageDataFromCommitStatus(project_config, task_config, org_config)
        task._init_task()
        with pytest.raises(
            DependencyLookupError,
            match="Could not find package version id in '2gp' commit status",
        ):
            task._run_task()

    @responses.activate
    def test_get_dependencies__version_not_found(self):
        responses.add(
            "GET",
            "https://salesforce/services/data/v49.0/tooling/query/",
            json={"records": []},
        )

        project_config = create_project_config(repo_commit="abcdef")
        project_config.keychain.set_service(
            "github",
            ServiceConfig(
                {
                    "username": "TestUser",
                    "password": "TestPass",
                    "email": "testuser@testdomain.com",
                }
            ),
        )
        task_config = TaskConfig({"options": {"context": "2gp"}})
        org_config = OrgConfig(
            {"instance_url": "https://salesforce", "access_token": "TOKEN"}, "test"
        )
        task = GetPackageDataFromCommitStatus(project_config, task_config, org_config)
        task._init_task()
        with pytest.raises(
            DependencyLookupError, match="Could not look up dependencies of 04t"
        ):
            task._get_dependencies("04t")



from cumulusci.tasks.salesforce import BaseSalesforceApiTask
from cumulusci.core.utils import process_list_of_pairs_dict_arg, process_bool_arg
from cumulusci.core.exceptions import SalesforceException


class InsertRecord(BaseSalesforceApiTask):
    task_docs = """
        For example:

        cci task run insert_record --org dev -o object PermissionSet -o values Name:HardDelete,PermissionsBulkApiHardDelete:true
    """

    task_options = {
        "object": {"description": "An sObject type to insert", "required": True},
        "values": {
            "description": "Field names and values in the format 'aa:bb,cc:dd', or a YAML dict in cumulusci.yml.",
            "required": True,
        },
        "tooling": {"description": "If True, use the Tooling API instead of REST API."},
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        self.values = process_list_of_pairs_dict_arg(self.options["values"])
        self.object = self.options["object"]
        self.use_tooling = process_bool_arg(self.options.get("tooling", False))

    def _run_task(self):
        api = self.sf if not self.use_tooling else self.tooling
        object_handler = getattr(api, self.object)

        rc = object_handler.create(self.values)
        if rc["success"]:
            self.logger.info(f"{self.object} record inserted: {rc['id']}")
        else:
            # this will probably never execute, due to simple_salesforce throwing
            # an exception, but just in case:
            raise SalesforceException(
                f"Could not insert {self.object} record : {rc['errors']}"
            )



# -*- coding: utf-8 -*-
"""Format and display CLI output.

Classes:
    CliTable: Pretty prints tabular data to stdout, via click.echo.
"""
import os
import click
import textwrap
from terminaltables import AsciiTable, SingleTable


CHECKMARK = click.style("✔" if os.name == "posix" else "+", fg="green")
CROSSMARK = click.style("✘" if os.name == "posix" else "-", fg="red")


class CliTable:
    """Format and print data to the command line in tabular form.
    Attributes:

    * INNER_BORDER: Boolean passed to terminaltables.inner_row_border. Defaults to True.
    * PICTOGRAM_TRUE: True boolean values are replaced with this string.
    * PICTOGRAM_FALSE = False boolean values are replaced with this string.

    Methods:
    * echo: Print the table data to stdout using click.echo()
    * pretty_table: Table drawn using Unicode drawing characters.
    * ascii_table: Table drawn using Ascii characters.
    """

    INNER_BORDER = True
    PICTOGRAM_TRUE = CHECKMARK
    PICTOGRAM_FALSE = " "

    def __init__(self, data, title=None, wrap_cols=None, bool_cols=None, dim_rows=None):
        """Constructor.
        Args:
            data: Required. List[List] of data to format, with the heading as 0-th member.
            title: String to use as the table's title.
            wrap_cols: List[str] of column names to wrap to max width.
            bool_cols: List[str] of columns containing booleans to stringify.
            dim_rows: List[int] of row indices to dim.
        """
        self._data = data
        self._header = data[0]
        self._title = title
        self._table = SingleTable(self._data, self._title)

        if wrap_cols:
            self._table_wrapper(self._table, wrap_cols)
        if bool_cols:
            for name in bool_cols:
                self.stringify_boolean_col(col_name=name)
        if dim_rows:
            self._dim_row_list(dim_rows)

    def _table_wrapper(self, table, wrap_cols):
        """Query for column width and wrap text"""
        for col_name in wrap_cols:
            index = self._get_index_for_col_name(col_name)
            width = abs(table.column_max_width(index))
            for row in table.table_data:
                row[index] = (
                    textwrap.fill(row[index], width) if row[index] is not None else ""
                )

    def stringify_boolean_col(self, col_name=None, true_str=None, false_str=None):
        """Replace booleans in the given column name with a string.
        Args:

        * col_name: str indicating which columns should be stringifed.
        * true_str: True values will be replaced with this string on posix systems.
        * false_str: False values will be replaced with this string on posix systems.
        """
        col_index = self._get_index_for_col_name(col_name)

        true_str = (
            click.style(true_str, fg="green") if true_str else self.PICTOGRAM_TRUE
        )
        false_str = (
            click.style(false_str, fg="red") if false_str else self.PICTOGRAM_FALSE
        )
        for row in self._table.table_data[1:]:
            row[col_index] = true_str if row[col_index] else false_str

    def echo(self, plain=False):
        """Print this table's data using click.echo().

        Automatically falls back to AsciiTable if there's an encoding error.
        """
        if plain or os.environ.get("TERM") == "dumb":
            table = self.ascii_table()
        else:
            table = str(self)
        click.echo(table)

    def __str__(self):
        try:
            return self.pretty_table()
        except UnicodeEncodeError:
            return self.ascii_table()

    def pretty_table(self):
        """Pretty prints a table."""
        self._table.inner_row_border = self.INNER_BORDER
        return self._table.table + "\n"

    def ascii_table(self):
        """Fallback for dumb terminals."""
        self.plain = AsciiTable(self._table.table_data, self._title)
        self.plain.inner_column_border = False
        self.plain.inner_row_border = False
        return self.plain.table

    def _get_index_for_col_name(self, col_name):
        return self._header.index(col_name)

    def _dim_row_list(self, dim_rows=None):
        """Given a list of integers, iterate over the table data and dim rows.

        Converts each value into a string.
        """
        for row_index in dim_rows:
            if row_index != 0:
                self._table.table_data[row_index] = [
                    self._dim_value(cell) for cell in self._table.table_data[row_index]
                ]

    def _dim_value(self, val):
        """Given an object, convert to a string and wrap using click.style."""
        val = str(val)

        # If a string has been wrapped, wrap each line to avoid dimming table borders
        dimmed_strs = [click.style(line, dim=True) for line in val.split("\n")]
        return "\n".join(dimmed_strs)


def _soql_table(results, truncated):
    if results:
        if truncated:
            assert results[-1] == truncated
            first_row = results[0]
            fake_row = {k: "" for k, v in first_row.items()}
            first_column = list(first_row)[0]
            fake_row[first_column] = truncated

            results[-1] = fake_row

        headings = list(results[0].keys())
        return str(
            CliTable(
                [headings] + [list(map(str, r.values())) for r in results],
                wrap_cols=headings,
            )
        )
    else:
        return str(CliTable([["No results"]]))


def _summarize(field):
    if field["referenceTo"]:
        allowed = field["referenceTo"]
    elif field["picklistValues"]:
        allowed = [value["value"] for value in field["picklistValues"]]
    else:
        allowed = field["type"]
    return (field["name"], allowed)


class SimpleSalesforceUIHelpers:
    def __init__(self, sf):
        self._sf = sf

    def query(self, query, format="table", include_deleted=False, max_rows=100):
        """Return the result of a Salesforce SOQL query.

        Arguments:

            * query -- the SOQL query to send to Salesforce, e.g.
                    SELECT Id FROM Lead WHERE Email = "waldo@somewhere.com"
            * include_deleted -- True if deleted records should be included
            * format -- one of these values:
                - "table" -- printable ASCII table (the default)
                - "obj" -- ordinary Python objects
                - "pprint" -- string in easily readable Python dict shape
                - "json" -- JSON
            * max_rows -- maximum rows to output, defaults to 100

        For example:
            query("select Name, Id from Account", format="pprint")
            contact_ids = query("select count(Id) from Contact", format="obj")
        """
        results = self._sf.query_all(query, include_deleted=include_deleted)["records"]

        if len(results) > max_rows:
            truncated = f"... truncated {len(results) - max_rows} rows"
            results = results[0:max_rows]
        else:
            truncated = False

        for result in results:
            if result.get("attributes"):
                del result["attributes"]

        if truncated:
            results.append(truncated)

        if format == "table":
            help_message = "Type help(query) to learn about other return formats or assigning the result."
            print(_soql_table(results, truncated))
            print()
            print(help_message)
            rc = None
        elif format == "obj":
            rc = results
        elif format == "pprint":
            from pprint import pprint

            pprint(results)
            rc = None
        elif format == "json":
            from json import dumps

            rc = dumps(results, indent=2)
        else:
            raise TypeError(f"Unknown format `{format}`")

        return rc

    def describe(self, sobj_name, detailed=False, format="pprint"):
        """Describe an sobject.

        Arguments:

            sobj_name - sobject name to describe. e.g. "Account", "Contact"
            detailed - set to `True` to get detailed information about object
            format -- one of these values:
            - "pprint" -- string in easily readable Python dict shape (default)
            - "obj" -- ordinary Python objects

        For example:

            >>> describe("Account")
            >>> data = describe("Account", detailed=True, format=obj)
        """
        from pprint import pprint

        data = getattr(self._sf, sobj_name).describe()

        if detailed:
            rc = data
        else:
            rc = dict(_summarize(field) for field in data["fields"])

        if format == "pprint":
            pprint(rc)
        elif format == "obj":
            return rc
        else:
            raise TypeError(f"Unknown format {format}")



from base64 import b64encode
from xml.sax.saxutils import escape
import contextlib
import html
import functools
import io
import logging
import os
import pathlib
import zipfile

from cumulusci.core.sfdx import sfdx
from cumulusci.utils import cd
from cumulusci.utils import inject_namespace
from cumulusci.utils import process_text_in_zipfile
from cumulusci.utils import strip_namespace
from cumulusci.utils import temporary_dir
from cumulusci.utils import tokenize_namespace
from cumulusci.utils import zip_clean_metaxml
from cumulusci.utils.xml import metadata_tree
from cumulusci.utils.ziputils import hash_zipfile_contents

INSTALLED_PACKAGE_PACKAGE_XML = """<?xml version="1.0" encoding="utf-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
  <types>
    <members>{namespace}</members>
    <name>InstalledPackage</name>
  </types>
<version>{version}</version>
</Package>"""

EMPTY_PACKAGE_XML = """<?xml version="1.0" encoding="utf-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
<version>{version}</version>
</Package>"""

FULL_NAME_PACKAGE_XML = """<?xml version="1.0" encoding="utf-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
<fullName>{}</fullName>
<version>{}</version>
</Package>"""

INSTALLED_PACKAGE = """<?xml version="1.0" encoding="UTF-8"?>
<InstalledPackage xmlns="http://soap.sforce.com/2006/04/metadata">
  <versionNumber>{}</versionNumber>
  <activateRSS>{}</activateRSS>
  <securityType>{}</securityType>
  {}
</InstalledPackage>"""

DEFAULT_LOGGER = logging.getLogger(__name__)


class BasePackageZipBuilder(object):
    def __init__(self):
        self._open_zip()

    def _open_zip(self):
        """Start a new, empty zipfile"""
        self.buffer = io.BytesIO()
        self.zf = zipfile.ZipFile(self.buffer, "w", zipfile.ZIP_DEFLATED)

    def _write_package_xml(self, package_xml):
        self.zf.writestr("package.xml", package_xml)

    def _write_file(self, path, content):
        self.zf.writestr(path, content)

    def as_bytes(self):
        fp = self.zf.fp
        self.zf.close()
        value = fp.getvalue()
        fp.close()
        return value

    def as_base64(self):
        return b64encode(self.as_bytes()).decode("utf-8")

    def as_hash(self):
        return hash_zipfile_contents(self.zf)

    def __call__(self):
        # for backwards compatibility
        return self.as_base64()


class MetadataPackageZipBuilder(BasePackageZipBuilder):
    """Build a package zip from a metadata folder."""

    def __init__(
        self,
        *,
        path=None,
        zf: zipfile.ZipFile = None,
        options=None,
        logger=None,
        name=None,
    ):
        self.options = options or {}
        self.logger = logger or DEFAULT_LOGGER
        if zf is not None:
            self.zf = zf
        elif path is not None:
            self._open_zip()
            with self._convert_sfdx_format(path, name) as path:
                self._add_files_to_package(path)
        else:
            self._open_zip()
        self._process()

    @classmethod
    def from_zipfile(cls, zf, *, options=None, logger=None):
        """Start with an existing zipfile rather than a filesystem folder."""
        return cls(zf=zf, options=options, logger=logger)

    @contextlib.contextmanager
    def _convert_sfdx_format(self, path, name):
        orig_path = path
        with contextlib.ExitStack() as stack:
            if not pathlib.Path(path, "package.xml").exists():
                self.logger.info("Converting from sfdx to mdapi format")
                path = stack.enter_context(temporary_dir(chdir=False))
                args = ["-r", str(orig_path), "-d", path]
                if name:
                    args += ["-n", name]
                sfdx(
                    "force:source:convert",
                    args=args,
                    capture_output=False,
                    check_return=True,
                )

            yield path

    def _add_files_to_package(self, path):
        for file_path in self._find_files_to_package(path):
            relpath = str(file_path.relative_to(path)).replace(os.sep, "/")
            self.zf.write(file_path, arcname=relpath)

    def _find_files_to_package(self, path):
        """Generator of paths to include in the package.

        Walks through all directories and files in path,
        filtering using _include_directory and _include_file
        """
        for root, dirs, files in os.walk(path):
            root_parts = pathlib.Path(root).relative_to(path).parts
            if self._include_directory(root_parts):
                for f in files:
                    if self._include_file(root_parts, f):
                        yield pathlib.Path(root, f)

    def _include_directory(self, root_parts):
        """Return boolean for whether this directory should be included in the package."""
        # include root
        if len(root_parts) == 0:
            return True

        # include top level only within lwc
        if root_parts[0] == "lwc" and len(root_parts) != 2:
            return False

        # include everything else
        return True

    def _include_file(self, root_parts, f):
        """Return boolean for whether this file should be included in the package."""
        if len(root_parts) == 2 and root_parts[0] == "lwc":
            # is file of lwc component directory
            lower_f = f.lower()
            return lower_f.endswith((".js", ".js-meta.xml", ".html", ".css", ".svg"))
        return True

    def _process(self):
        # We have to close the existing zipfile and reopen it before processing;
        # otherwise we hit a bug in Windows where ZipInfo objects have the wrong path separators.
        fp = self.zf.fp
        self.zf.close()
        self.zf = zipfile.ZipFile(fp, "r")

        self._process_namespace_tokens()
        self._clean_meta_xml()
        self._bundle_staticresources()
        self._remove_feature_parameters()

    def _process_namespace_tokens(self):
        zipf = self.zf
        if self.options.get("namespace_tokenize"):
            self.logger.info(
                f"Tokenizing namespace prefix {self.options['namespace_tokenize']}__"
            )
            zipf = process_text_in_zipfile(
                zipf,
                functools.partial(
                    tokenize_namespace,
                    namespace=self.options["namespace_tokenize"],
                    logger=self.logger,
                ),
            )
        if self.options.get("namespace_inject"):
            managed = not self.options.get("unmanaged", True)
            if managed:
                self.logger.info(
                    "Replacing namespace tokens from metadata with namespace prefix  "
                    f"{self.options['namespace_inject']}__"
                )
            else:
                self.logger.info(
                    "Stripping namespace tokens from metadata for unmanaged deployment"
                )
            zipf = process_text_in_zipfile(
                zipf,
                functools.partial(
                    inject_namespace,
                    namespace=self.options["namespace_inject"],
                    managed=managed,
                    namespaced_org=self.options.get("namespaced_org", False),
                    logger=self.logger,
                ),
            )
        if self.options.get("namespace_strip"):
            self.logger.info("Stripping namespace tokens from metadata")
            zipf = process_text_in_zipfile(
                zipf,
                functools.partial(
                    strip_namespace,
                    namespace=self.options["namespace_strip"],
                    logger=self.logger,
                ),
            )
        self.zf = zipf

    def _clean_meta_xml(self):
        if not self.options.get("clean_meta_xml", True):
            return
        self.logger.info(
            "Cleaning meta.xml files of packageVersion elements for deploy"
        )
        zf = zip_clean_metaxml(self.zf)
        self.zf.close()
        self.zf = zf

    def _bundle_staticresources(self):
        relpath = self.options.get("static_resource_path")
        if not relpath or not os.path.exists(relpath):
            return
        path = os.path.realpath(relpath)

        # Copy existing files to new zipfile
        zip_dest = zipfile.ZipFile(io.BytesIO(), "w", zipfile.ZIP_DEFLATED)
        for name in self.zf.namelist():
            if name == "package.xml":
                package_xml = self.zf.open(name)
            else:
                content = self.zf.read(name)
                zip_dest.writestr(name, content)

        # Build static resource bundles and add to package
        with temporary_dir():
            os.mkdir("staticresources")
            bundles = []
            for name in os.listdir(path):
                bundle_relpath = os.path.join(relpath, name)
                bundle_path = os.path.join(path, name)
                if not os.path.isdir(bundle_path):
                    continue
                self.logger.info(
                    "Zipping {} to add to staticresources".format(bundle_relpath)
                )

                # Add resource-meta.xml file
                meta_name = "{}.resource-meta.xml".format(name)
                meta_path = os.path.join(path, meta_name)
                with open(meta_path, "rb") as f:
                    zip_dest.writestr("staticresources/{}".format(meta_name), f.read())

                # Add bundle
                zip_path = os.path.join("staticresources", "{}.resource".format(name))
                with open(zip_path, "wb") as bundle_fp:
                    bundle_zip = zipfile.ZipFile(bundle_fp, "w", zipfile.ZIP_DEFLATED)
                    with cd(bundle_path):
                        for root, dirs, files in os.walk("."):
                            for f in files:
                                resource_file = os.path.join(root, f)
                                bundle_zip.write(resource_file)
                    bundle_zip.close()
                zip_dest.write(zip_path)
                bundles.append(name)

        # Update package.xml
        Package = metadata_tree.parse(package_xml)
        sections = Package.findall("types", name="StaticResource")
        section = sections[0] if sections else None
        if not section:
            section = Package.append("types")
            section.append("name", text="StaticResource")
        for name in bundles:
            section.insert_before(section.find("name"), tag="members", text=name)
        package_xml = Package.tostring(xml_declaration=True)
        zip_dest.writestr("package.xml", package_xml)

        self.zf.close()
        self.zf = zip_dest

    def _remove_feature_parameters(self):
        # Remove feature parameters from unlocked packages only
        if self.options.get("package_type") != "Unlocked":
            return

        # Copy existing files to new zipfile
        package_xml = None
        zip_dest = zipfile.ZipFile(io.BytesIO(), "w", zipfile.ZIP_DEFLATED)
        for name in self.zf.namelist():
            if name == "package.xml":
                package_xml = self.zf.open(name)
            elif name.startswith("featureParameters/"):
                # skip feature parameters
                self.logger.info(f"Skipping {name} in unlocked package")
                continue
            else:
                content = self.zf.read(name)
                zip_dest.writestr(name, content)

        # Remove from package.xml
        if package_xml is not None:
            Package = metadata_tree.parse(package_xml)
            for mdtype in (
                "FeatureParameterInteger",
                "FeatureParameterString",
                "FeatureParameterBoolean",
            ):
                section = Package.find("types", name=mdtype)
                if section is not None:
                    Package.remove(section)
            package_xml = Package.tostring(xml_declaration=True)
            zip_dest.writestr("package.xml", package_xml)

        self.zf.close()
        self.zf = zip_dest


class CreatePackageZipBuilder(BasePackageZipBuilder):
    def __init__(self, name, api_version):
        if not name:
            raise ValueError("You must provide a name to create a package")
        if not api_version:
            raise ValueError("You must provide an api_version to create a package")
        self.name = name
        self.api_version = api_version

        self._open_zip()
        self._populate_zip()

    def _populate_zip(self):
        package_xml = FULL_NAME_PACKAGE_XML.format(escape(self.name), self.api_version)
        self._write_package_xml(package_xml)


class InstallPackageZipBuilder(BasePackageZipBuilder):
    api_version = "43.0"

    def __init__(
        self, namespace, version, activateRSS=False, password=None, securityType="FULL"
    ):
        if not namespace:
            raise ValueError("You must provide a namespace to install a package")
        if not version:
            raise ValueError("You must provide a version to install a package")
        self.namespace = namespace
        self.version = version
        self.activateRSS = activateRSS
        self.password = password
        self.securityType = securityType

        self._open_zip()
        self._populate_zip()

    def _populate_zip(self):
        package_xml = INSTALLED_PACKAGE_PACKAGE_XML.format(
            namespace=self.namespace, version=self.api_version
        )
        self._write_package_xml(package_xml)

        activateRSS = "true" if self.activateRSS else "false"
        password = (
            "<password>{}</password>".format(html.escape(self.password))
            if self.password
            else ""
        )
        installed_package = INSTALLED_PACKAGE.format(
            self.version, activateRSS, self.securityType, password
        )
        self._write_file(
            "installedPackages/{}.installedPackage".format(self.namespace),
            installed_package,
        )


class DestructiveChangesZipBuilder(BasePackageZipBuilder):
    def __init__(self, destructive_changes, version):
        self.destructive_changes = destructive_changes
        self.version = version

        self._open_zip()
        self._populate_zip()

    def _populate_zip(self):
        self._write_package_xml(EMPTY_PACKAGE_XML.format(version=self.version))
        self._write_file("destructiveChanges.xml", self.destructive_changes)


class UninstallPackageZipBuilder(DestructiveChangesZipBuilder):
    def __init__(self, namespace, version):
        if not namespace:
            raise ValueError("You must provide a namespace to install a package")
        self.namespace = namespace
        self.version = version
        self.destructive_changes = INSTALLED_PACKAGE_PACKAGE_XML.format(
            namespace=self.namespace, version=self.version
        )

        self._open_zip()
        self._populate_zip()



from collections import defaultdict
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.metadata_etl import MetadataSingleEntityTransformTask
from cumulusci.utils.xml.metadata_tree import MetadataElement
from cumulusci.core.utils import process_list_arg, process_bool_arg


class SetFieldHelpText(MetadataSingleEntityTransformTask):
    entity = "CustomObject"
    task_options = {
        "fields": {
            "description": "List of object fields to affect, in Object__c.Field__c form.",
            "required": True,
        },
        "overwrite": {
            "description": "If set to True, overwrite any differing Help Text found on the field. "
            "By default, Help Text is set only if it is blank."
        },
        **MetadataSingleEntityTransformTask.task_options,
    }

    def _init_options(self, kwargs):
        self.task_config.options["api_names"] = "dummy"
        super()._init_options(kwargs)

        self.options["overwrite"] = process_bool_arg(
            self.options.get("overwrite", False)
        )

        if (
            not isinstance(self.options.get("fields"), list)
            or len(self.options["fields"]) == 0
        ):
            raise TaskOptionsError(
                "Please populate the fields field with a list of dictionaries containing at minimum one entry with an 'api_name' and 'help_text' keys"
            )
        if not all(["api_name" in entry for entry in self.options["fields"]]):
            raise TaskOptionsError(
                "The 'api_name' key is required on all entry values."
            )
        if not all(["help_text" in entry for entry in self.options["fields"]]):
            raise TaskOptionsError(
                "The 'help_text' key is required on all entry values to declare what help text value to insert."
            )
        self.api_name_list = defaultdict(list)
        for entry in process_list_arg(self.options["fields"]):
            try:
                obj, field = entry["api_name"].split(".")

                self.api_name_list[self._inject_namespace(obj)].append(
                    (self._inject_namespace(field), entry["help_text"])
                )
            except ValueError:
                raise TaskOptionsError(
                    f"api_name {entry} is not a valid Object.Field reference"
                )

        self.api_names = set(self.api_name_list.keys())

    def _transform_entity(self, metadata: MetadataElement, api_name: str):
        for field, help_text in self.api_name_list[api_name]:
            self._modify_help_text(metadata, api_name, field, help_text)
        return metadata

    def _modify_help_text(
        self,
        metadata: MetadataElement,
        api_name: str,
        custom_field: str,
        help_text: str,
    ):
        # Locate the <fields> entry for this field entry.
        field = metadata.find("fields", fullName=custom_field)
        if not field:
            raise TaskOptionsError(
                f"The field {api_name}.{custom_field} was not found."
            )
        try:
            if (
                field.inlineHelpText.text == ""
                or field.inlineHelpText.text == help_text
                or self.options["overwrite"]
            ):
                field.inlineHelpText.text = help_text
            else:
                self.logger.warning(
                    f"Help text for field {api_name} has a different value. "
                    "Set the overwrite option to True to overwrite this field help text."
                )
        except AttributeError:
            field.append("inlineHelpText", text=help_text)



import pytest

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.salesforce.tests.util import create_task
from cumulusci.tasks.metadata_etl.help_text import SetFieldHelpText
from cumulusci.utils.xml import metadata_tree

# Custom Object with 2 custom fields with all elements present
OBJECT_XML = b"""<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <fields>
        <fullName>Foo__c</fullName>
        <inlineHelpText>Foo</inlineHelpText>
    </fields>
    <fields>
        <fullName>Bar__c</fullName>
        <inlineHelpText>Bar</inlineHelpText>
    </fields>
</CustomObject>
"""

# Additional Custom Object with 2 custom fields with all elements present
OBJECT_XML_2 = b"""<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <fields>
        <fullName>Foo__c</fullName>
        <inlineHelpText>Foo</inlineHelpText>
    </fields>
    <fields>
        <fullName>Bar__c</fullName>
        <inlineHelpText>Does something.</inlineHelpText>
    </fields>
</CustomObject>
"""

# Custom Object with 2 custom fields with one missing inlineHelpText
OBJECT_XML_3 = b"""<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <fields>
        <fullName>Foo__c</fullName>
    </fields>
    <fields>
        <fullName>Bar__c</fullName>
        <inlineHelpText>Does something.</inlineHelpText>
    </fields>
</CustomObject>
"""

# Custom Object with 2 custom fields both missing inlineHelpText
OBJECT_XML_4 = b"""<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <fields>
        <fullName>Foo__c</fullName>
        <defaultValue>false</defaultValue>
        <description>Indicates that the Contact is allowed to receive information protected by FERPA and other privacy laws, regulations, and policies.</description>
        <externalId>false</externalId>
        <label>FERPA Approved</label>
        <trackHistory>false</trackHistory>
        <trackTrending>false</trackTrending>
        <type>Checkbox</type>
    </fields>
    <fields>
        <fullName>Bar__c</fullName>
        <defaultValue>false</defaultValue>
        <description>Indicates that the Contact is allowed to receive information protected by HIPPA and other privacy laws, regulations, and policies.</description>
        <externalId>false</externalId>
        <label>HIPPA Approved</label>
        <trackHistory>false</trackHistory>
        <trackTrending>false</trackTrending>
        <type>Checkbox</type>
    </fields>
</CustomObject>
"""

STANDARD_OBJECT_XML = b"""<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <fields>
        <fullName>Foo</fullName>
        <defaultValue>false</defaultValue>
        <description>Indicates that the Contact is allowed to receive information protected by FERPA and other privacy laws, regulations, and policies.</description>
        <externalId>false</externalId>
        <label>FERPA Approved</label>
        <trackHistory>false</trackHistory>
        <trackTrending>false</trackTrending>
        <type>Checkbox</type>
    </fields>
    <fields>
        <fullName>Bar</fullName>
        <defaultValue>false</defaultValue>
        <description>Indicates that the Contact is allowed to receive information protected by HIPPA and other privacy laws, regulations, and policies.</description>
        <externalId>false</externalId>
        <inlineHelpText>Foo</inlineHelpText>
        <label>HIPPA Approved</label>
        <trackHistory>false</trackHistory>
        <trackTrending>false</trackTrending>
        <type>Checkbox</type>
    </fields>
</CustomObject>
"""


class TestAddPicklistValues:
    def test_add_single_object_help_text(self):
        task = create_task(
            SetFieldHelpText,
            {
                "api_version": "47.0",
                "overwrite": True,
                "fields": [{"api_name": "MyObject.Foo__c", "help_text": "foo"}],
            },
        )

        # Validate that the first sObject has one picklist changed
        tree = metadata_tree.fromstring(OBJECT_XML)
        result = task._transform_entity(tree, "MyObject")
        test_elem = result.find("fields", fullName="Foo__c")

        assert test_elem is not None
        assert test_elem.inlineHelpText.text == "foo"
        # Validate that the sObject alters only the custom field listed
        test_elem = result.find("fields", fullName="Bar__c")
        assert test_elem is not None
        assert test_elem.inlineHelpText.text == "Bar"

    def test_add_single_object_help_text__same_value(self):
        task = create_task(
            SetFieldHelpText,
            {
                "api_version": "47.0",
                "overwrite": True,
                "fields": [{"api_name": "MyObject.Foo__c", "help_text": "foo"}],
            },
        )

        tree = metadata_tree.fromstring(OBJECT_XML)
        test_elem = tree.find("fields", fullName="Foo__c")
        test_elem.inlineHelpText.text = "foo"
        assert task._transform_entity(tree, "MyObject") is tree

        assert test_elem.inlineHelpText.text == "foo"

    def test_add_single_object_help_text__blank_value(self):
        task = create_task(
            SetFieldHelpText,
            {
                "api_version": "47.0",
                "overwrite": True,
                "fields": [{"api_name": "MyObject.Foo__c", "help_text": "foo"}],
            },
        )

        tree = metadata_tree.fromstring(OBJECT_XML)
        test_elem = tree.find("fields", fullName="Foo__c")
        test_elem.inlineHelpText.text = ""
        assert task._transform_entity(tree, "MyObject") is tree

        assert test_elem.inlineHelpText.text == "foo"

    def test_add_multi_object_help_text(self):
        task = create_task(
            SetFieldHelpText,
            {
                "api_version": "47.0",
                "overwrite": True,
                "fields": [
                    {"api_name": "MyObject.Foo__c", "help_text": "foo"},
                    {"api_name": "MyObject2.Bar__c", "help_text": "bar"},
                ],
            },
        )

        # Validate that the first sObject has one picklist changed
        tree = metadata_tree.fromstring(OBJECT_XML)
        result = task._transform_entity(tree, "MyObject")
        test_elem = result.find("fields", fullName="Foo__c")

        assert test_elem is not None
        assert test_elem.inlineHelpText.text == "foo"
        # Validate that the first sObject alters only the custom field listed
        test_elem = result.find("fields", fullName="Bar__c")
        assert test_elem is not None
        assert test_elem.inlineHelpText.text == "Bar"
        tree = metadata_tree.fromstring(OBJECT_XML_2)
        result = task._transform_entity(tree, "MyObject2")
        test_elem = result.find("fields", fullName="Bar__c")

        assert test_elem is not None
        assert test_elem.inlineHelpText.text == "bar"

        test_elem = result.find("fields", fullName="Foo__c")
        # Validate that the second sObject alters only the custom field listed: Tobias__c
        assert test_elem is not None
        assert test_elem.inlineHelpText.text == "Foo"

    def test_add_single_object_multi_help_text(self):
        task = create_task(
            SetFieldHelpText,
            {
                "api_version": "47.0",
                "overwrite": True,
                "fields": [
                    {"api_name": "MyObject.Foo__c", "help_text": "foo"},
                    {"api_name": "MyObject.Bar__c", "help_text": "bar"},
                ],
            },
        )

        tree = metadata_tree.fromstring(OBJECT_XML)
        result = task._transform_entity(tree, "MyObject")
        # Validate that the first sObject has one picklist changed
        test_elem = result.find("fields", fullName="Foo__c")
        assert test_elem is not None
        assert test_elem.inlineHelpText.text == "foo"
        # Validate that the first sObject alters only the custom field listed
        test_elem = result.find("fields", fullName="Bar__c")
        assert test_elem is not None
        assert test_elem.inlineHelpText.text == "bar"

    def test_add_single_object_no_help_text(self):
        task = create_task(
            SetFieldHelpText,
            {
                "api_version": "47.0",
                "overwrite": True,
                "fields": [
                    {"api_name": "MyObject3.Foo__c", "help_text": "foo"},
                    {"api_name": "MyObject3.Bar__c", "help_text": "bar"},
                ],
            },
        )

        tree = metadata_tree.fromstring(OBJECT_XML_3)
        result = task._transform_entity(tree, "MyObject3")
        # Validate that the first sObject has one picklist changed
        test_elem = result.find("fields", fullName="Foo__c")
        assert test_elem is not None
        assert test_elem.inlineHelpText.text == "foo"
        # Validate that the first sObject alters only the custom field listed
        test_elem = result.find("fields", fullName="Bar__c")
        assert test_elem is not None
        assert test_elem.inlineHelpText.text == "bar"

    def test_add_single_object_multi_field_no_help_text(self):
        task = create_task(
            SetFieldHelpText,
            {
                "api_version": "47.0",
                "overwrite": True,
                "fields": [
                    {"api_name": "MyObject4.Foo__c", "help_text": "foo"},
                    {"api_name": "MyObject4.Bar__c", "help_text": "bar"},
                ],
            },
        )

        tree = metadata_tree.fromstring(OBJECT_XML_4)
        result = task._transform_entity(tree, "MyObject4")
        # Validate that the first sObject has one picklist changed
        test_elem = result.find("fields", fullName="Foo__c")
        assert test_elem is not None
        assert test_elem.inlineHelpText.text == "foo"
        # Validate that the first sObject alters only the custom field listed
        test_elem = result.find("fields", fullName="Bar__c")
        assert test_elem is not None
        assert test_elem.inlineHelpText.text == "bar"

    def test_raises_for_empty_fields(self):
        with pytest.raises(TaskOptionsError):
            task = create_task(SetFieldHelpText, {"api_version": "47.0", "fields": []})

            tree = metadata_tree.fromstring(OBJECT_XML)
            task._transform_entity(tree, "MyObject")

    def test_raises_for_non_list_fields(self):
        with pytest.raises(TaskOptionsError):
            task = create_task(
                SetFieldHelpText,
                {
                    "api_version": "47.0",
                    "fields": {"api_name": "MyObject.Foo__c", "help_text": "foo"},
                },
            )

            tree = metadata_tree.fromstring(OBJECT_XML)
            task._transform_entity(tree, "MyObject")

    def test_raises_no_fields_help_text(self):
        with pytest.raises(TaskOptionsError):
            task = create_task(
                SetFieldHelpText,
                {"api_version": "47.0", "fields": [{"api": "MyObject.Foo__c"}]},
            )

            tree = metadata_tree.fromstring(OBJECT_XML)
            task._transform_entity(tree, "MyObject")

    def test_raises_no_fields_api(self):
        with pytest.raises(TaskOptionsError):
            task = create_task(
                SetFieldHelpText,
                {"api_version": "47.0", "fields": [{"help_text": "help"}]},
            )

            tree = metadata_tree.fromstring(OBJECT_XML)
            task._transform_entity(tree, "MyObject")

    def test_sets_helptext_for_standard_field_fields(self):
        task = create_task(
            SetFieldHelpText,
            {
                "api_version": "47.0",
                "overwrite": True,
                "fields": [
                    {"api_name": "MyObject.Foo", "help_text": "foo"},
                    {"api_name": "MyObject.Bar", "help_text": "bar"},
                ],
            },
        )

        tree = metadata_tree.fromstring(STANDARD_OBJECT_XML)
        result = task._transform_entity(tree, "MyObject")
        test_elem = result.find("fields", fullName="Foo")
        assert test_elem is not None
        assert test_elem.inlineHelpText.text == "foo"

        test_elem = result.find("fields", fullName="Bar")
        assert test_elem is not None
        assert test_elem.inlineHelpText.text == "bar"

    def test_raises_for_no_help_text_field(self):
        with pytest.raises(TaskOptionsError):
            task = create_task(
                SetFieldHelpText,
                {
                    "api_version": "47.0",
                    "fields": [{"api_name": "MyObject.Foo", "bar": "buster_name"}],
                },
            )

            tree = metadata_tree.fromstring(STANDARD_OBJECT_XML)
            task._transform_entity(tree, "MyObject")

    def test_raises_invalid_object(self):
        with pytest.raises(TaskOptionsError):
            task = create_task(
                SetFieldHelpText,
                {
                    "api_version": "48.0",
                    "fields": [{"api_name": "Test.c.Test__c", "help_text": "help"}],
                },
            )

            tree = metadata_tree.fromstring(OBJECT_XML)
            task._transform_entity(tree, "MyObject")

    def test_raises_missing_object(self):
        with pytest.raises(TaskOptionsError):
            task = create_task(
                SetFieldHelpText,
                {
                    "api_version": "48.0",
                    "fields": [
                        {"api_name": "MyObject.sherlock__c", "help_text": "help"}
                    ],
                },
            )

            tree = metadata_tree.fromstring(OBJECT_XML)
            task._transform_entity(tree, "MyObject")

    def test_overwrite_false(self):
        task = create_task(
            SetFieldHelpText,
            {
                "api_version": "47.0",
                "overwrite": False,
                "fields": [{"api_name": "MyObject.Foo__c", "help_text": "foo"}],
            },
        )

        # Validate that the first sObject has one picklist changed
        tree = metadata_tree.fromstring(OBJECT_XML)
        result = task._transform_entity(tree, "MyObject")
        test_elem = result.find("fields", fullName="Foo__c")

        assert test_elem is not None
        assert test_elem.inlineHelpText.text == "Foo"



import http.client

from github3 import GitHubError
import github3.exceptions

from cumulusci.core.exceptions import GithubApiNotFoundError
from cumulusci.core.utils import process_bool_arg
from cumulusci.tasks.github.base import BaseGithubTask


class MergeBranchOld(BaseGithubTask):
    task_options = {
        "commit": {
            "description": "The commit to merge into feature branches.  Defaults to the current head commit."
        },
        "source_branch": {
            "description": "The source branch to merge from.  Defaults to project__git__default_branch."
        },
        "branch_prefix": {
            "description": "The prefix of branches that should receive the merge.  Defaults to project__git__prefix_feature"
        },
        "children_only": {
            "description": "If True, merge will only be done to child branches.  This assumes source branch is a parent feature branch.  Defaults to False"
        },
    }

    def _init_options(self, kwargs):
        super(MergeBranchOld, self)._init_options(kwargs)

        if "commit" not in self.options:
            self.options["commit"] = self.project_config.repo_commit
        if "branch_prefix" not in self.options:
            self.options[
                "branch_prefix"
            ] = self.project_config.project__git__prefix_feature
        if "source_branch" not in self.options:
            self.options[
                "source_branch"
            ] = self.project_config.project__git__default_branch
        self.options["children_only"] = process_bool_arg(
            self.options.get("children_only", False)
        )

    def _run_task(self):
        self.repo = self.get_repo()

        self._validate_branch()
        self._get_existing_prs()
        branch_tree = self._get_branch_tree()
        self._merge_branches(branch_tree)

    def _validate_branch(self):
        try:
            self.repo.branch(self.options["source_branch"])
        except github3.exceptions.NotFoundError:
            message = f"Branch {self.options['source_branch']} not found"
            self.logger.error(message)
            raise GithubApiNotFoundError(message)

    def _get_existing_prs(self):
        # Get existing pull requests targeting a target branch
        self.existing_prs = []
        for pr in self.repo.pull_requests(state="open"):
            if (
                pr.base.ref.startswith(self.options["branch_prefix"])
                and pr.head.ref == self.options["source_branch"]
            ):
                self.existing_prs.append(pr.base.ref)

    def _get_branch_tree(self):
        branches, branches_dict = self._get_list_and_dict_of_branches()
        children, parents = self._get_parent_and_child_branches(branches)
        return self._construct_branch_tree(branches, branches_dict, children, parents)

    def _get_list_and_dict_of_branches(self):
        """Returns a list and dict of branches that match the given branch_prefix"""
        branches = []
        branches_dict = {}
        for branch in self.repo.branches():
            if branch.name == self.options["source_branch"]:
                if not self.options["children_only"]:
                    self.logger.debug(
                        f"Skipping branch {branch.name}: is source branch"
                    )
                    branches_dict[branch.name] = branch
                    continue
            if not branch.name.startswith(self.options["branch_prefix"]):
                if not self.options["children_only"]:
                    self.logger.debug(
                        f"Skipping branch {branch.name}: does not match prefix {self.options['branch_prefix']}"
                    )
                # The following line isn't included in coverage
                # due to behavior of the CPython peephole optimizer,
                # see https://bitbucket.org/ned/coveragepy/issues/198/continue-marked-as-not-covered
                continue  # pragma: no cover
            branches.append(branch)
            branches_dict[branch.name] = branch

        return branches, branches_dict

    def _get_parent_and_child_branches(self, branches):
        possible_children = []
        possible_parents = []
        for branch in branches:
            parts = branch.name.replace(self.options["branch_prefix"], "", 1).split(
                "__", 1
            )
            if len(parts) == 2:
                possible_children.append(parts)
            else:
                possible_parents.append(branch.name)

        parents = {}
        children = []
        for possible_child in possible_children:
            parent = f"{self.options['branch_prefix']}{possible_child[0]}"
            if parent in possible_parents:
                child = "__".join(possible_child)
                child = self.options["branch_prefix"] + child
                if parent not in parents:
                    parents[parent] = []
                parents[parent].append(child)
                children.append(child)

        return children, parents

    def _construct_branch_tree(self, branches, branches_dict, children, parents):
        """Build a branch tree list with parent/child branches"""
        branch_tree = []
        for branch in branches:
            if branch.name in children:
                # Skip child branches
                continue
            if (
                self.options["children_only"]
                and branch.name != self.options["source_branch"]
            ):
                # If merging to children only, skip any branches other than source
                continue
            branch_item = {"branch": branch, "children": []}
            for child in parents.get(branch.name, []):
                branch_item["children"].append(branches_dict[child])
            branch_tree.append(branch_item)

        return branch_tree

    def _merge_branches(self, branch_tree):
        # Process merge on all branches
        for branch_item in branch_tree:
            if self.options["children_only"]:
                if branch_item["children"]:
                    self.logger.info(
                        f"Performing merge from parent branch {self.options['source_branch']} to children"
                    )
                else:
                    self.logger.info(
                        f"No children found for branch {self.options['source_branch']}"
                    )
                    continue
                for child in branch_item["children"]:
                    self._merge(
                        branch=child.name,
                        source=self.options["source_branch"],
                        commit=self.options["commit"],
                        children=[],
                    )
            else:
                self._merge(
                    branch=branch_item["branch"].name,
                    source=self.options["source_branch"],
                    commit=self.options["commit"],
                    children=branch_item["children"],
                )

    def _merge(self, branch, source, commit, children=None):
        if not children:
            children = []
        branch_type = "branch"
        if children:
            branch_type = "parent branch"
        if self.options["children_only"]:
            branch_type = "child branch"

        compare = self.repo.compare_commits(branch, commit)
        if not compare or not compare.files:
            self.logger.info(f"Skipping {branch_type} {branch}: no file diffs found")
            return

        try:
            self.repo.merge(branch, commit)
            self.logger.info(
                f"Merged {compare.behind_by} commits into {branch_type} {branch}"
            )
            if children and not self.options["children_only"]:
                self.logger.info("  Skipping merge into the following child branches:")
                for child in children:
                    self.logger.info(f"    {child.name}")

        except GitHubError as e:
            if e.code != http.client.CONFLICT:
                raise

            if branch in self.existing_prs:
                self.logger.info(
                    f"Merge conflict on {branch_type} {branch}: merge PR already exists"
                )
                return

            pull = self.repo.create_pull(
                title=f"Merge {source} into {branch}",
                base=branch,
                head=source,
                body="This pull request was automatically generated because "
                "an automated merge hit a merge conflict",
            )

            self.logger.info(
                f"Merge conflict on {branch_type} {branch}: created pull request #{pull.number}"
            )


class MergeBranch(BaseGithubTask):
    task_docs = """
    Merges the most recent commit on the current branch into other branches depending on the value of source_branch.

    If source_branch is a branch that does not start with the specified branch_prefix, then the commit will be
    merged to all branches that begin with branch_prefix and are not themselves child branches (i.e. branches don't contain '__' in their name).

    If source_branch begins with branch_prefix, then the commit is merged to all child branches of source_branch.
    """
    task_options = {
        "commit": {
            "description": "The commit to merge into feature branches.  Defaults to the current head commit."
        },
        "source_branch": {
            "description": "The source branch to merge from.  Defaults to project__git__default_branch."
        },
        "branch_prefix": {
            "description": "A list of prefixes of branches that should receive the merge.  Defaults to project__git__prefix_feature"
        },
        "update_future_releases": {
            "description": "If source_branch is a release branch, then merge all future release branches that exist. Defaults to False."
        },
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)

        if "commit" not in self.options:
            self.options["commit"] = self.project_config.repo_commit
        if "branch_prefix" not in self.options:
            self.options[
                "branch_prefix"
            ] = self.project_config.project__git__prefix_feature
        if "source_branch" not in self.options:
            self.options[
                "source_branch"
            ] = self.project_config.project__git__default_branch
        self.options["update_future_releases"] = process_bool_arg(
            self.options.get("update_future_releases", False)
        )

    def _run_task(self):
        self.repo = self.get_repo()
        self.source_branch_is_default = (
            self.options["source_branch"]
            == self.project_config.project__git__default_branch
        )
        self._validate_source_branch()
        self._get_existing_prs()

        branches_to_merge = self._get_branches_to_merge()

        for branch in branches_to_merge:
            self._merge(
                branch.name,
                self.options["source_branch"],
                self.options["commit"],
            )

    def _validate_source_branch(self):
        """Validates that the source branch exists in the repository"""
        try:
            self.repo.branch(self.options["source_branch"])
        except github3.exceptions.NotFoundError:
            message = f"Branch {self.options['source_branch']} not found"
            raise GithubApiNotFoundError(message)

    def _get_existing_prs(self):
        """Get existing pull requests from the source branch
        to other branches that are candidates for merging."""
        self.existing_prs = []
        for pr in self.repo.pull_requests(state="open"):
            if (
                pr.base.ref.startswith(self.options["branch_prefix"])
                and pr.head.ref == self.options["source_branch"]
            ):
                self.existing_prs.append(pr.base.ref)

    def _get_branches_to_merge(self):
        """
        If source_branch is the default branch, we
        gather all branches with branch_prefix that are not child branches.

        If source_branch is not the default branch, we gather
        all branches with branch_prefix that are direct descendents of source_branch.

        If update_future_releases is True, and source_branch is a release branch
        then we also collect all future release branches.
        """
        child_branches = []
        main_descendents = []
        release_branches = []
        for branch in self.repo.branches():
            if (
                self._is_release_branch(self.options["source_branch"])
                and self.options["update_future_releases"]
                and self._is_future_release_branch(branch.name)
            ):
                release_branches.append(branch)
                continue

            if branch.name == self.options["source_branch"]:
                self.logger.debug(f"Skipping branch {branch.name}: is source branch")
                continue
            elif not branch.name.startswith(self.options["branch_prefix"]):
                self.logger.debug(
                    f"Skipping branch {branch.name}: does not match prefix '{self.options['branch_prefix']}'"
                )
                continue
            elif (
                not self.options["source_branch"].startswith(
                    self.options["branch_prefix"]
                )
                and "__" not in branch.name
            ):
                main_descendents.append(branch)
            elif self._is_source_branch_direct_descendent(branch):
                child_branches.append(branch)
            else:
                self.logger.debug(
                    f"Skipping branch {branch.name}: is not a direct descendent of {self.options['source_branch']}"
                )

        to_merge = []
        if child_branches:
            self.logger.debug(
                f"Found child branches to update: {[branch.name for branch in child_branches]}"
            )
            to_merge = child_branches
        elif self.options["source_branch"].startswith(self.options["branch_prefix"]):
            self.logger.debug(
                f"No children found for branch {self.options['source_branch']}"
            )

        if release_branches:
            self.logger.debug(
                f"Found future release branches to update: {[branch.name for branch in release_branches]}"
            )
            to_merge = to_merge + release_branches

        if main_descendents:
            self.logger.debug(
                f"Found descendents of {self.options['source_branch']} to update: {[branch.name for branch in main_descendents]}"
            )
            to_merge = to_merge + main_descendents

        return to_merge

    def _merge(self, branch_name, source, commit):
        """Attempt to merge a commit from source to branch with branch_name"""
        compare = self.repo.compare_commits(branch_name, commit)
        if not compare or not compare.files:
            self.logger.info(f"Skipping branch {branch_name}: no file diffs found")
            return

        try:
            self.repo.merge(branch_name, commit)
            self.logger.info(
                f"Merged {compare.behind_by} commits into branch: {branch_name}"
            )

        except GitHubError as e:
            if e.code != http.client.CONFLICT:
                raise

            if branch_name in self.existing_prs:
                self.logger.info(
                    f"Merge conflict on branch {branch_name}: merge PR already exists"
                )
                return

            pull = self.repo.create_pull(
                title=f"Merge {source} into {branch_name}",
                base=branch_name,
                head=source,
                body="This pull request was automatically generated because "
                "an automated merge hit a merge conflict",
            )

            self.logger.info(
                f"Merge conflict on branch {branch_name}: created pull request #{pull.number}"
            )

    def _is_source_branch_direct_descendent(self, branch):
        """Returns True if branch is a direct descendent of the source branch"""
        source_dunder_count = self.options["source_branch"].count("__")
        return (
            branch.name.startswith(f"{self.options['source_branch']}__")
            and branch.name.count("__") == source_dunder_count + 1
        )

    def _is_future_release_branch(self, branch_name):
        return (
            self._is_release_branch(branch_name)
            and branch_name != self.options["source_branch"]
            and self._get_release_num(branch_name)
            > self._get_release_num(self.options["source_branch"])
        )

    def _is_release_branch(self, branch_name):
        """A release branch begins with the given prefix"""
        prefix = self.options["branch_prefix"]
        if not branch_name.startswith(prefix):
            return False
        parts = branch_name[len(prefix) :].split("__")
        return len(parts) == 1 and parts[0].isdigit()

    def _get_release_num(self, release_branch_name):
        """Given a release branch, returns an integer that
        corresponds to the release number for that branch"""
        return int(release_branch_name.split(self.options["branch_prefix"])[1])



import pytest

from io import BytesIO
from pathlib import Path
from cumulusci.utils.xml.metadata_tree import METADATA_NAMESPACE, parse, fromstring

standard_xml = f"""<Data xmlns='{METADATA_NAMESPACE}'>
                <foo>Foo</foo>
                <foo>Foo2</foo>
                <bar><name>Bar1</name><label>Label1</label></bar>
                <bar><name>Bar2</name><label>Label2</label></bar>
                <text>Baz</text>
            </Data>"""


class TestMetadataTree:
    def test_insertions(self):
        Data = fromstring(
            f"""<Data xmlns='{METADATA_NAMESPACE}'><Foo>foo1</Foo><Foo>foo2</Foo></Data>"""
        )

        assert Data.Foo[0].text == "foo1"
        assert Data.Foo[1].text == "foo2"

        foo4 = Data.append("Foo", "foo4")
        assert Data.Foo[2].text == "foo4"
        assert Data.find(tag="Foo", text="foo4") == foo4

        Data.insert_before(foo4, tag="Foo", text="foo3")
        assert Data.Foo[2].text == "foo3"
        Data.insert_after(foo4, tag="Foo", text="foo5")
        assert Data.Foo[4].text == "foo5"

    def test_text(self):
        Data = fromstring(
            f"""<Data xmlns='{METADATA_NAMESPACE}'>
                <foo>Foo</foo>
                <text>Bar</text>
                <text>
                    <bar>Bar2</bar>
                </text>
            </Data>"""
        )
        assert isinstance(Data.foo.text, str)
        assert Data.foo.text == "Foo"
        assert isinstance(Data.text.text, str)
        assert Data.text.text.lower() == "bar"
        assert Data.text[1].bar.text.lower() == "bar2"

    def test_append_goes_in_the_middle(self):
        Data = fromstring(
            f"""<Data xmlns='{METADATA_NAMESPACE}'>
                <foo>Foo</foo>
                <bar>Bar</bar>
            </Data>"""
        )
        Data.append(tag="foo", text="Foo2")
        assert Data.foo[1].text == "Foo2"
        Data.append(tag="foo", text="Foo3")
        assert Data.foo[2].text == "Foo3"
        Data.append(tag="bar", text="Bar2")
        assert Data.bar[1].text == "Bar2"

    def test_whitespace(self):
        from cumulusci.tasks.metadata import tests

        path = (
            Path(tests.__file__).parent
            / "package_metadata/namespaced_report_folder/package.xml"
        )
        with open(path) as f:
            raw = f.read()

        # get rid of whitespace to see if we can replace it faithfully
        raw_flattened = raw.replace("    ", "").replace("\n", "")
        Package = parse(BytesIO(raw_flattened.encode("utf-8")))
        x = Package.tostring(xml_declaration=True).strip()
        assert x == raw

    def test_pretty_printing(self):
        Data = fromstring(f"<Data xmlns='{METADATA_NAMESPACE}'/>")
        Data.append("A", "AA")
        B = Data.append("B")
        B.append("C", "CC")
        Data.append("A", "AB")

        assert (
            Data.tostring(xml_declaration=True)
            == """<?xml version="1.0" encoding="UTF-8"?>
<Data xmlns="http://soap.sforce.com/2006/04/metadata">
    <A>AA</A>
    <A>AB</A>
    <B>
        <C>CC</C>
    </B>
</Data>
"""
        )

    def test_remove(self):
        Data = fromstring(f"<Data xmlns='{METADATA_NAMESPACE}'/>")
        Data.append("A", "AA")
        B = Data.append("B")
        B.append("C", "CC")
        Data.append("A", "AB")

        assert Data.A[0].text == "AA"
        assert Data.B.C.text == "CC"
        assert Data.A[1].text == "AB"

        Data.remove(Data.A[0])
        Data.remove(Data.B)
        Data.remove(Data.A)
        assert Data.tostring() == f'<Data xmlns="{METADATA_NAMESPACE}"/>\n'

    def test_error_handling(self):
        Data = fromstring(f"<Data xmlns='{METADATA_NAMESPACE}'><Foo/></Data>\n")
        Data.Foo
        with pytest.raises(AttributeError) as e:
            assert Data.Bar
        assert "not found in" in str(e.value)

    def test_getattr_getitem(self):
        Data = fromstring(f"<Data xmlns='{METADATA_NAMESPACE}'><Foo/></Data>")
        assert Data.Foo == Data["Foo"]

    def test_missing_text(self):
        Data = fromstring(f"<Data xmlns='{METADATA_NAMESPACE}'><Foo/></Data>")
        with pytest.raises(AttributeError) as e:
            Data.text
        assert "not found in" in str(e.value)

    def test_pathlib_support(self):
        from cumulusci.tasks.metadata import tests

        path = (
            Path(tests.__file__).parent
            / "package_metadata/namespaced_report_folder/package.xml"
        )
        tree = parse(path)
        assert tree

    def test_getitem_error(self):
        Data = fromstring(f"<Data xmlns='{METADATA_NAMESPACE}'><Foo/></Data>")
        with pytest.raises(TypeError):
            Data[None]

    def test_matching(self):
        Data = fromstring(standard_xml)
        assert Data.find("foo").text == Data.findall("foo")[0].text == "Foo"
        assert Data.find("foo", text="Foo2").text == "Foo2"

        assert Data.findall("foo", text="Foo2") == [Data.find("foo", text="Foo2")]

        assert Data.find("foo", name="xyzzy") is None
        assert Data.findall("foo", name="xyzzy") == []
        assert Data.find("foo", text="xyzzy") is None
        assert Data.findall("foo", text="xyzzy") == []
        assert Data.find("bar", name="Bar1").label.text == "Label1"
        assert Data.findall("bar", name="Bar1") == [Data.find("bar", name="Bar1")]
        assert Data.find("bar", name="Bar2").label.text == "Label2"
        assert Data.findall("bar", name="Bar2") == [Data.find("bar", name="Bar2")]
        assert Data.find("bar", name="xyzzy") is None
        assert Data.find("text").text == "Baz"
        assert Data.find("text", text="Baz").text == "Baz"

    def test_equality(self):
        Data = fromstring(standard_xml)
        assert Data.foo == Data.foo[0]

    def test_iteration(self):
        Data = fromstring(standard_xml)

        for foo in Data.foo:
            assert foo.tag == "foo"

        for bar in Data.bar:
            assert bar.tag == "bar"

        # this one might be slightly non-intutive but its a
        # consequence of how the system works
        for child in Data.bar[0]:
            assert child.tag == "bar"

    def test_repr(self):
        Data = fromstring(standard_xml)
        for foo in Data.foo:
            assert repr(foo)

        for bar in Data.bar:
            assert repr(bar)

        Data = fromstring(f"<Data xmlns='{METADATA_NAMESPACE}'></Data>")
        assert repr(Data) == "<Data></Data> element"

    # you'll need to temporarily disable Typeguard to test this properly
    def test_getitem_type_checking(self):
        Data = fromstring(standard_xml)
        with pytest.raises(TypeError):
            Data.foo[None]

    def test_multiple_namespaces(self):
        xml = """
<CustomMetadata xmlns="http://soap.sforce.com/2006/04/metadata" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xsd="http://www.w3.org/2001/XMLSchema">
    <label>Account MD Isolation Rollup</label>
    <protected>false</protected>
    <values>
        <field>dlrs__Active__c</field>
        <value xsi:type="xsd:boolean">true</value>
    </values>
</CustomMetadata>""".strip()
        CustomMetadata = fromstring(xml)
        assert xml.strip() == CustomMetadata.tostring().strip()



from cumulusci.tasks.preflight.settings import CheckSettingsValue
from cumulusci.tasks.salesforce.tests.util import create_task

from simple_salesforce.exceptions import SalesforceMalformedRequest

import pytest
import responses


JSON_RESPONSE = {
    "records": [{"IntVal": 3, "FloatVal": 3.0, "BoolVal": True, "StringVal": "foo"}],
    "done": True,
    "totalSize": 1,
}


@responses.activate
@pytest.mark.parametrize(
    "settings_field,value,outcome",
    [
        ("IntVal", 3, True),
        ("FloatVal", 3.0, True),
        ("BoolVal", "true", True),
        ("StringVal", "foo", True),
        ("StringVal", "bad", False),
    ],
)
def test_check_settings(settings_field, value, outcome):
    responses.add(
        "GET",
        f"https://test.salesforce.com/services/data/v49.0/tooling/query/?q=SELECT+{settings_field}+FROM+ChatterSettings",
        json=JSON_RESPONSE,
    )
    task = create_task(
        CheckSettingsValue,
        {
            "settings_type": "ChatterSettings",
            "settings_field": settings_field,
            "value": value,
        },
    )

    task()

    assert task.return_values is outcome


@responses.activate
def test_check_settings__no_settings():
    responses.add(
        "GET",
        "https://test.salesforce.com/services/data/v49.0/tooling/query/?q=SELECT+Foo+FROM+ChatterSettings",
        json={"records": []},
    )
    task = create_task(
        CheckSettingsValue,
        {
            "settings_type": "ChatterSettings",
            "settings_field": "Foo",
            "value": True,
        },
    )

    task()

    assert task.return_values is False


@responses.activate
def test_check_settings__failure():
    responses.add(
        "GET",
        status=400,
        url="https://test.salesforce.com/services/data/v49.0/tooling/query/?q=SELECT+Test+FROM+NoSettings",
        json={},
    )
    task = create_task(
        CheckSettingsValue,
        {
            "settings_type": "NoSettings",
            "settings_field": "Test",
            "value": True,
            "treat_missing_as_failure": True,
        },
    )

    task()

    assert task.return_values is False


@responses.activate
def test_check_settings__exception():
    responses.add(
        "GET",
        status=400,
        url="https://test.salesforce.com/services/data/v49.0/tooling/query/?q=SELECT+Test+FROM+NoSettings",
        json={},
    )
    task = create_task(
        CheckSettingsValue,
        {
            "settings_type": "NoSettings",
            "settings_field": "Test",
            "value": True,
        },
    )

    with pytest.raises(SalesforceMalformedRequest):
        task()

    assert task.return_values is False



from unittest import mock
import os
import unittest

from cumulusci.core.config import UniversalConfig
from cumulusci.core.config import BaseProjectConfig
from cumulusci.core.config import TaskConfig
from cumulusci.tasks.metaxml import UpdateApi
from cumulusci.tasks.metaxml import UpdateDependencies
from cumulusci.utils import temporary_dir


class TestUpdateApi(unittest.TestCase):
    def test_run_task(self):
        with temporary_dir() as d:
            os.mkdir(".git")
            os.mkdir("src")
            meta_xml_path = os.path.join(d, "src", "test-meta.xml")
            with open(meta_xml_path, "w") as f:
                f.write(
                    """<?xml version="1.0" encoding="UTF-8"?>
<ApexClass xmlns="http://soap.sforce.com/2006/04/metadata">
    <apiVersion>1.0</apiVersion>
</ApexClass>
"""
                )

            project_config = BaseProjectConfig(
                UniversalConfig(), config={"noyaml": True}
            )
            task_config = TaskConfig({"options": {"version": "43.0"}})
            task = UpdateApi(project_config, task_config)
            task()

            with open(meta_xml_path, "r") as f:
                result = f.read()
            self.assertEqual(
                """<?xml version="1.0" encoding="UTF-8"?>
<ApexClass xmlns="http://soap.sforce.com/2006/04/metadata">
    <apiVersion>43.0</apiVersion>
</ApexClass>
""",
                result,
            )


class TestUpdateDependencies(unittest.TestCase):
    def test_run_task(self):
        with temporary_dir() as d:
            os.mkdir(".git")
            os.mkdir("src")
            meta_xml_path = os.path.join(d, "src", "test-meta.xml")
            with open(meta_xml_path, "w") as f:
                f.write(
                    """<?xml version="1.0" encoding="UTF-8"?>
<ApexClass xmlns="http://soap.sforce.com/2006/04/metadata">
    <packageVersions>
        <namespace>npsp</namespace>
        <majorNumber>1</majorNumber>
        <minorNumber>0</minorNumber>
    </packageVersions>
    <packageVersions>
        <namespace>npe01</namespace>
        <majorNumber>1</majorNumber>
        <minorNumber>0</minorNumber>
    </packageVersions>
</ApexClass>
"""
                )

            project_config = BaseProjectConfig(
                UniversalConfig(), config={"noyaml": True}
            )
            project_config.get_static_dependencies = mock.Mock(
                return_value=[
                    {
                        "namespace": "npsp",
                        "version": "3.0",
                        "dependencies": [{"namespace": "npe01", "version": "1.1"}],
                    }
                ]
            )
            task_config = TaskConfig()
            task = UpdateDependencies(project_config, task_config)
            task()

            with open(meta_xml_path, "r") as f:
                result = f.read()
            self.assertEqual(
                """<?xml version="1.0" encoding="UTF-8"?>
<ApexClass xmlns="http://soap.sforce.com/2006/04/metadata">
    <packageVersions>
        <namespace>npsp</namespace>
        <majorNumber>3</majorNumber>
        <minorNumber>0</minorNumber>
    </packageVersions>
    <packageVersions>
        <namespace>npe01</namespace>
        <majorNumber>1</majorNumber>
        <minorNumber>1</minorNumber>
    </packageVersions>
</ApexClass>
""",
                result,
            )



from cumulusci.tasks.salesforce.BaseSalesforceApiTask import BaseSalesforceApiTask
from cumulusci.core.utils import process_list_arg


class CreatePermissionSet(BaseSalesforceApiTask):
    task_options = {
        "api_name": {
            "description": "API name of generated Permission Set",
            "required": True,
        },
        "label": {"description": "Label of generated Permission Set"},
        "user_permissions": {
            "description": "List of User Permissions to include in the Permission Set.",
            "required": True,
        },
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)

        self.options["user_permissions"] = process_list_arg(
            self.options["user_permissions"]
        )

    def _run_task(self):
        result = self.sf.PermissionSet.create(
            {
                "Name": self.options["api_name"],
                "Label": self.options.get("label") or self.options["api_name"],
                **{perm: True for perm in self.options["user_permissions"]},
            }
        )
        self.sf.PermissionSetAssignment.create(
            {"AssigneeId": self.org_config.user_id, "PermissionSetId": result["id"]}
        )



import io
import json
import unittest
from unittest import mock

import responses

from cumulusci.core.exceptions import BulkDataException
from cumulusci.tasks.bulkdata.step import (
    download_file,
    DataOperationType,
    DataOperationStatus,
    DataOperationResult,
    DataOperationJobResult,
    BulkJobMixin,
    BulkApiQueryOperation,
    BulkApiDmlOperation,
    RestApiQueryOperation,
    RestApiDmlOperation,
    DataApi,
    get_query_operation,
    get_dml_operation,
)
from cumulusci.tasks.bulkdata.load import LoadData
from cumulusci.tests.util import mock_describe_calls
from cumulusci.tasks.bulkdata.tests.utils import _make_task


BULK_BATCH_RESPONSE = """<root xmlns="http://ns">
<batch>
    <state>{first_state}</state>
    <stateMessage>{first_message}</stateMessage>
</batch>
<batch>
    <state>{second_state}</state>
    <stateMessage>{second_message}</stateMessage>
</batch>
</root>"""


class TestDownloadFile(unittest.TestCase):
    @responses.activate
    def test_download_file(self):
        url = "https://example.com"
        bulk_mock = mock.Mock()
        bulk_mock.headers.return_value = {}

        responses.add(method="GET", url=url, body=b"TEST\xe2\x80\x94")
        with download_file(url, bulk_mock) as f:
            # make sure it was decoded as utf-8
            assert f.read() == "TEST\u2014"


class TestBulkDataJobTaskMixin(unittest.TestCase):
    @responses.activate
    def test_job_state_from_batches(self):
        mixin = BulkJobMixin()
        mixin.bulk = mock.Mock()
        mixin.bulk.endpoint = "https://example.com"
        mixin.bulk.headers.return_value = {"HEADER": "test"}
        mixin._parse_job_state = mock.Mock()

        responses.add(
            "GET",
            "https://example.com/job/JOB/batch",
            adding_headers=mixin.bulk.headers.return_value,
            body="TEST",
        )
        assert (
            mixin._job_state_from_batches("JOB") == mixin._parse_job_state.return_value
        )
        mixin._parse_job_state.assert_called_once_with(b"TEST")

    def test_parse_job_state(self):
        mixin = BulkJobMixin()
        mixin.bulk = mock.Mock()
        mixin.bulk.jobNS = "http://ns"

        assert (
            mixin._parse_job_state(
                BULK_BATCH_RESPONSE.format(
                    **{
                        "first_state": "Not Processed",
                        "first_message": "Test",
                        "second_state": "Completed",
                        "second_message": "",
                    }
                )
            )
            == DataOperationJobResult(DataOperationStatus.ABORTED, [], 0, 0)
        )

        assert (
            mixin._parse_job_state(
                BULK_BATCH_RESPONSE.format(
                    **{
                        "first_state": "InProgress",
                        "first_message": "Test",
                        "second_state": "Completed",
                        "second_message": "",
                    }
                )
            )
            == DataOperationJobResult(DataOperationStatus.IN_PROGRESS, [], 0, 0)
        )

        assert mixin._parse_job_state(
            BULK_BATCH_RESPONSE.format(
                **{
                    "first_state": "Failed",
                    "first_message": "Bad",
                    "second_state": "Failed",
                    "second_message": "Worse",
                }
            )
        ) == DataOperationJobResult(
            DataOperationStatus.JOB_FAILURE, ["Bad", "Worse"], 0, 0
        )

        assert (
            mixin._parse_job_state(
                BULK_BATCH_RESPONSE.format(
                    **{
                        "first_state": "Completed",
                        "first_message": "Test",
                        "second_state": "Completed",
                        "second_message": "",
                    }
                )
            )
            == DataOperationJobResult(DataOperationStatus.SUCCESS, [], 0, 0)
        )

        assert mixin._parse_job_state(
            '<root xmlns="http://ns">'
            "  <batch><state>Completed</state></batch>"
            "  <numberRecordsFailed>200</numberRecordsFailed>"
            "</root>"
        ) == DataOperationJobResult(DataOperationStatus.ROW_FAILURE, [], 0, 200)

        assert mixin._parse_job_state(
            '<root xmlns="http://ns">'
            "  <batch><state>Completed</state></batch>"
            "  <numberRecordsFailed>200</numberRecordsFailed>"
            "  <numberRecordsProcessed>10</numberRecordsProcessed>"
            "</root>"
        ) == DataOperationJobResult(DataOperationStatus.ROW_FAILURE, [], 10, 200)

    @mock.patch("time.sleep")
    def test_wait_for_job(self, sleep_patch):
        mixin = BulkJobMixin()

        mixin.bulk = mock.Mock()
        mixin.bulk.job_status.return_value = {
            "numberBatchesCompleted": 1,
            "numberBatchesTotal": 1,
        }
        mixin._job_state_from_batches = mock.Mock(
            side_effect=[
                DataOperationJobResult(DataOperationStatus.IN_PROGRESS, [], 0, 0),
                DataOperationJobResult(DataOperationStatus.SUCCESS, [], 0, 0),
            ]
        )
        mixin.logger = mock.Mock()

        result = mixin._wait_for_job("750000000000000")
        mixin._job_state_from_batches.assert_has_calls(
            [mock.call("750000000000000"), mock.call("750000000000000")]
        )
        assert result.status is DataOperationStatus.SUCCESS

    def test_wait_for_job__failed(self):
        mixin = BulkJobMixin()

        mixin.bulk = mock.Mock()
        mixin.bulk.job_status.return_value = {
            "numberBatchesCompleted": 1,
            "numberBatchesTotal": 1,
        }
        mixin._job_state_from_batches = mock.Mock(
            return_value=DataOperationJobResult(
                DataOperationStatus.JOB_FAILURE, ["Test1", "Test2"], 0, 0
            )
        )
        mixin.logger = mock.Mock()

        result = mixin._wait_for_job("750000000000000")
        mixin._job_state_from_batches.assert_called_once_with("750000000000000")
        assert result.status is DataOperationStatus.JOB_FAILURE

    def test_wait_for_job__logs_state_messages(self):
        mixin = BulkJobMixin()

        mixin.bulk = mock.Mock()
        mixin.bulk.job_status.return_value = {
            "numberBatchesCompleted": 1,
            "numberBatchesTotal": 1,
        }
        mixin._job_state_from_batches = mock.Mock(
            return_value=DataOperationJobResult(
                DataOperationStatus.JOB_FAILURE, ["Test1", "Test2"], 0, 0
            )
        )
        mixin.logger = mock.Mock()

        mixin._wait_for_job("750000000000000")
        mixin.logger.error.assert_any_call("Batch failure message: Test1")
        mixin.logger.error.assert_any_call("Batch failure message: Test2")


class TestBulkApiQueryOperation(unittest.TestCase):
    def test_query(self):
        context = mock.Mock()
        query = BulkApiQueryOperation(
            sobject="Contact",
            api_options={},
            context=context,
            query="SELECT Id FROM Contact",
        )
        query._wait_for_job = mock.Mock()
        query._wait_for_job.return_value = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 0, 0
        )

        query.query()

        assert query.job_result.status is DataOperationStatus.SUCCESS

        context.bulk.create_query_job.assert_called_once_with(
            "Contact", contentType="CSV"
        )
        context.bulk.query.assert_called_once_with(
            context.bulk.create_query_job.return_value, "SELECT Id FROM Contact"
        )
        query._wait_for_job.assert_called_once_with(
            context.bulk.create_query_job.return_value
        )
        context.bulk.close_job.assert_called_once_with(
            context.bulk.create_query_job.return_value
        )

    def test_query__contextmanager(self):
        context = mock.Mock()
        query = BulkApiQueryOperation(
            sobject="Contact",
            api_options={},
            context=context,
            query="SELECT Id FROM Contact",
        )
        query._wait_for_job = mock.Mock()
        query._wait_for_job.return_value = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 0, 0
        )

        with query:
            assert query.job_result.status is DataOperationStatus.SUCCESS

    def test_query__failure(self):
        context = mock.Mock()
        query = BulkApiQueryOperation(
            sobject="Contact",
            api_options={},
            context=context,
            query="SELECT Id FROM Contact",
        )
        query._wait_for_job = mock.Mock()
        query._wait_for_job.return_value = DataOperationJobResult(
            DataOperationStatus.JOB_FAILURE, [], 0, 0
        )

        query.query()

        assert query.job_result.status is DataOperationStatus.JOB_FAILURE

    @mock.patch("cumulusci.tasks.bulkdata.step.download_file")
    def test_get_results(self, download_mock):
        context = mock.Mock()
        context.bulk.endpoint = "https://test"
        context.bulk.create_query_job.return_value = "JOB"
        context.bulk.query.return_value = "BATCH"
        context.bulk.get_query_batch_result_ids.return_value = ["RESULT"]

        download_mock.return_value = io.StringIO(
            """Id
003000000000001
003000000000002
003000000000003"""
        )
        query = BulkApiQueryOperation(
            sobject="Contact",
            api_options={},
            context=context,
            query="SELECT Id FROM Contact",
        )
        query._wait_for_job = mock.Mock()
        query._wait_for_job.return_value = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 0, 0
        )
        query.query()

        results = list(query.get_results())

        context.bulk.get_query_batch_result_ids.assert_called_once_with(
            "BATCH", job_id="JOB"
        )
        download_mock.assert_called_once_with(
            "https://test/job/JOB/batch/BATCH/result/RESULT", context.bulk
        )

        assert list(results) == [
            ["003000000000001"],
            ["003000000000002"],
            ["003000000000003"],
        ]

    @mock.patch("cumulusci.tasks.bulkdata.step.download_file")
    def test_get_results__no_results(self, download_mock):
        context = mock.Mock()
        context.bulk.endpoint = "https://test"
        context.bulk.create_query_job.return_value = "JOB"
        context.bulk.query.return_value = "BATCH"
        context.bulk.get_query_batch_result_ids.return_value = ["RESULT"]

        download_mock.return_value = io.StringIO("Records not found for this query")
        query = BulkApiQueryOperation(
            sobject="Contact",
            api_options={},
            context=context,
            query="SELECT Id FROM Contact",
        )
        query._wait_for_job = mock.Mock()
        query._wait_for_job.return_value = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 0, 0
        )
        query.query()

        results = list(query.get_results())

        context.bulk.get_query_batch_result_ids.assert_called_once_with(
            "BATCH", job_id="JOB"
        )
        download_mock.assert_called_once_with(
            "https://test/job/JOB/batch/BATCH/result/RESULT", context.bulk
        )

        assert list(results) == []


class TestBulkApiDmlOperation(unittest.TestCase):
    def test_start(self):
        context = mock.Mock()
        context.bulk.create_job.return_value = "JOB"

        step = BulkApiDmlOperation(
            sobject="Contact",
            operation=DataOperationType.INSERT,
            api_options={},
            context=context,
            fields=["LastName"],
        )

        step.start()

        context.bulk.create_job.assert_called_once_with(
            "Contact", "insert", contentType="CSV", concurrency="Parallel"
        )
        assert step.job_id == "JOB"

    def test_end(self):
        context = mock.Mock()
        context.bulk.create_job.return_value = "JOB"

        step = BulkApiDmlOperation(
            sobject="Contact",
            operation=DataOperationType.INSERT,
            api_options={},
            context=context,
            fields=["LastName"],
        )
        step._wait_for_job = mock.Mock()
        step._wait_for_job.return_value = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 0, 0
        )
        step.job_id = "JOB"

        step.end()

        context.bulk.close_job.assert_called_once_with("JOB")
        step._wait_for_job.assert_called_once_with("JOB")
        assert step.job_result.status is DataOperationStatus.SUCCESS

    def test_end__failed(self):
        context = mock.Mock()
        context.bulk.create_job.return_value = "JOB"

        step = BulkApiDmlOperation(
            sobject="Contact",
            operation=DataOperationType.INSERT,
            api_options={},
            context=context,
            fields=["LastName"],
        )
        step._wait_for_job = mock.Mock()
        step._wait_for_job.return_value = DataOperationJobResult(
            DataOperationStatus.JOB_FAILURE, [], 0, 0
        )
        step.job_id = "JOB"

        step.end()

        context.bulk.close_job.assert_called_once_with("JOB")
        step._wait_for_job.assert_called_once_with("JOB")
        assert step.job_result.status is DataOperationStatus.JOB_FAILURE

    def test_contextmanager(self):
        context = mock.Mock()
        context.bulk.create_job.return_value = "JOB"

        step = BulkApiDmlOperation(
            sobject="Contact",
            operation=DataOperationType.INSERT,
            api_options={},
            context=context,
            fields=["LastName"],
        )
        step._wait_for_job = mock.Mock()
        step._wait_for_job.return_value = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 0, 0
        )
        step.job_id = "JOB"

        with step:
            pass

        context.bulk.create_job.assert_called_once_with(
            "Contact", "insert", contentType="CSV", concurrency="Parallel"
        )
        assert step.job_id == "JOB"

        context.bulk.close_job.assert_called_once_with("JOB")
        step._wait_for_job.assert_called_once_with("JOB")
        assert step.job_result.status is DataOperationStatus.SUCCESS

    def test_serialize_csv_record(self):
        context = mock.Mock()
        step = BulkApiDmlOperation(
            sobject="Contact",
            operation=DataOperationType.INSERT,
            api_options={"batch_size": 2},
            context=context,
            fields=["Id", "FirstName", "LastName"],
        )

        serialized = step._serialize_csv_record(step.fields)
        assert serialized == b"Id,FirstName,LastName\r\n"

        record = ["1", "Bob", "Ross"]
        serialized = step._serialize_csv_record(record)
        assert serialized == b"1,Bob,Ross\r\n"

        record = ["col1", "multiline\ncol2"]
        serialized = step._serialize_csv_record(record)
        assert serialized == b'col1,"multiline\ncol2"\r\n'

    def test_batch(self):
        context = mock.Mock()

        step = BulkApiDmlOperation(
            sobject="Contact",
            operation=DataOperationType.INSERT,
            api_options={"batch_size": 2},
            context=context,
            fields=["LastName"],
        )

        records = iter([["Test"], ["Test2"], ["Test3"]])
        results = list(step._batch(records, n=2))

        assert len(results) == 2
        assert list(results[0]) == [
            "LastName\r\n".encode("utf-8"),
            "Test\r\n".encode("utf-8"),
            "Test2\r\n".encode("utf-8"),
        ]
        assert list(results[1]) == [
            "LastName\r\n".encode("utf-8"),
            "Test3\r\n".encode("utf-8"),
        ]

    def test_batch__character_limit(self):
        context = mock.Mock()

        step = BulkApiDmlOperation(
            sobject="Contact",
            operation=DataOperationType.INSERT,
            api_options={"batch_size": 2},
            context=context,
            fields=["LastName"],
        )

        records = [["Test"], ["Test2"], ["Test3"]]

        csv_rows = [step._serialize_csv_record(step.fields)]
        for r in records:
            csv_rows.append(step._serialize_csv_record(r))

        char_limit = sum([len(r) for r in csv_rows]) - 1

        # Ask for batches of three, but we
        # should get batches of 2 back
        results = list(step._batch(iter(records), n=3, char_limit=char_limit))

        assert len(results) == 2
        assert list(results[0]) == [
            "LastName\r\n".encode("utf-8"),
            "Test\r\n".encode("utf-8"),
            "Test2\r\n".encode("utf-8"),
        ]
        assert list(results[1]) == [
            "LastName\r\n".encode("utf-8"),
            "Test3\r\n".encode("utf-8"),
        ]

    @mock.patch("cumulusci.tasks.bulkdata.step.download_file")
    def test_get_results(self, download_mock):
        context = mock.Mock()
        context.bulk.endpoint = "https://test"
        download_mock.side_effect = [
            io.StringIO(
                """id,success,created,error
003000000000001,true,true,
003000000000002,true,true,"""
            ),
            io.StringIO(
                """id,success,created,error
003000000000003,false,false,error"""
            ),
        ]

        step = BulkApiDmlOperation(
            sobject="Contact",
            operation=DataOperationType.INSERT,
            api_options={},
            context=context,
            fields=["LastName"],
        )
        step.job_id = "JOB"
        step.batch_ids = ["BATCH1", "BATCH2"]

        results = step.get_results()

        assert list(results) == [
            DataOperationResult("003000000000001", True, None),
            DataOperationResult("003000000000002", True, None),
            DataOperationResult(None, False, "error"),
        ]
        download_mock.assert_has_calls(
            [
                mock.call("https://test/job/JOB/batch/BATCH1/result", context.bulk),
                mock.call("https://test/job/JOB/batch/BATCH2/result", context.bulk),
            ]
        )

    @mock.patch("cumulusci.tasks.bulkdata.step.download_file")
    def test_get_results__failure(self, download_mock):
        context = mock.Mock()
        context.bulk.endpoint = "https://test"
        download_mock.return_value.side_effect = Exception

        step = BulkApiDmlOperation(
            sobject="Contact",
            operation=DataOperationType.INSERT,
            api_options={},
            context=context,
            fields=["LastName"],
        )
        step.job_id = "JOB"
        step.batch_ids = ["BATCH1", "BATCH2"]

        with self.assertRaises(BulkDataException):
            list(step.get_results())

    @mock.patch("cumulusci.tasks.bulkdata.step.download_file")
    def test_end_to_end(self, download_mock):
        context = mock.Mock()
        context.bulk.endpoint = "https://test"
        context.bulk.create_job.return_value = "JOB"
        context.bulk.post_batch.side_effect = ["BATCH1", "BATCH2"]
        download_mock.return_value = io.StringIO(
            """id,success,created,error
003000000000001,true,true,
003000000000002,true,true,
003000000000003,false,false,error"""
        )

        step = BulkApiDmlOperation(
            sobject="Contact",
            operation=DataOperationType.INSERT,
            api_options={},
            context=context,
            fields=["LastName"],
        )
        step._wait_for_job = mock.Mock()
        step._wait_for_job.return_value = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 0, 0
        )

        step.start()
        step.load_records(iter([["Test"], ["Test2"], ["Test3"]]))
        step.end()

        assert step.job_result.status is DataOperationStatus.SUCCESS
        results = step.get_results()

        assert list(results) == [
            DataOperationResult("003000000000001", True, None),
            DataOperationResult("003000000000002", True, None),
            DataOperationResult(None, False, "error"),
        ]


class TestRestApiQueryOperation:
    def test_query(self):
        context = mock.Mock()
        context.sf.query.return_value = {
            "totalSize": 2,
            "done": True,
            "records": [
                {
                    "Id": "003000000000001",
                    "LastName": "Narvaez",
                    "Email": "wayne@example.com",
                },
                {"Id": "003000000000002", "LastName": "De Vries", "Email": None},
            ],
        }

        query_op = RestApiQueryOperation(
            sobject="Contact",
            fields=["Id", "LastName", "Email"],
            api_options={},
            context=context,
            query="SELECT Id, LastName,  Email FROM Contact",
        )

        query_op.query()

        assert query_op.job_result == DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 2, 0
        )
        assert list(query_op.get_results()) == [
            ["003000000000001", "Narvaez", "wayne@example.com"],
            ["003000000000002", "De Vries", ""],
        ]

    def test_query_batches(self):
        context = mock.Mock()
        context.sf.query.return_value = {
            "totalSize": 2,
            "done": False,
            "records": [
                {
                    "Id": "003000000000001",
                    "LastName": "Narvaez",
                    "Email": "wayne@example.com",
                }
            ],
            "nextRecordsUrl": "test",
        }

        context.sf.query_more.return_value = {
            "totalSize": 2,
            "done": True,
            "records": [
                {"Id": "003000000000002", "LastName": "De Vries", "Email": None}
            ],
        }

        query_op = RestApiQueryOperation(
            sobject="Contact",
            fields=["Id", "LastName", "Email"],
            api_options={},
            context=context,
            query="SELECT Id, LastName,  Email FROM Contact",
        )

        query_op.query()

        assert query_op.job_result == DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 2, 0
        )
        assert list(query_op.get_results()) == [
            ["003000000000001", "Narvaez", "wayne@example.com"],
            ["003000000000002", "De Vries", ""],
        ]


class TestRestApiDmlOperation:
    @responses.activate
    def test_insert_dml_operation(self):
        mock_describe_calls()
        task = _make_task(
            LoadData,
            {
                "options": {
                    "database_url": "sqlite:///test.db",
                    "mapping": "mapping.yml",
                }
            },
        )
        task.project_config.project__package__api_version = "48.0"
        task._init_task()

        responses.add(
            responses.POST,
            url="https://example.com/services/data/v48.0/composite/sobjects",
            json=[
                {"id": "003000000000001", "success": True},
                {"id": "003000000000002", "success": True},
            ],
            status=200,
        )
        responses.add(
            responses.POST,
            url="https://example.com/services/data/v48.0/composite/sobjects",
            json=[{"id": "003000000000003", "success": True}],
            status=200,
        )

        recs = [["Fred", "Narvaez"], [None, "De Vries"], ["Hiroko", "Aito"]]

        dml_op = RestApiDmlOperation(
            sobject="Contact",
            operation=DataOperationType.INSERT,
            api_options={"batch_size": 2},
            context=task,
            fields=["FirstName", "LastName"],
        )

        dml_op.start()
        dml_op.load_records(iter(recs))
        dml_op.end()

        assert dml_op.job_result == DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 3, 0
        )
        assert list(dml_op.get_results()) == [
            DataOperationResult("003000000000001", True, ""),
            DataOperationResult("003000000000002", True, ""),
            DataOperationResult("003000000000003", True, ""),
        ]

    @responses.activate
    def test_insert_dml_operation__boolean_conversion(self):
        mock_describe_calls()
        task = _make_task(
            LoadData,
            {
                "options": {
                    "database_url": "sqlite:///test.db",
                    "mapping": "mapping.yml",
                }
            },
        )
        task.project_config.project__package__api_version = "48.0"
        task._init_task()

        responses.add(
            responses.POST,
            url="https://example.com/services/data/v48.0/composite/sobjects",
            json=[
                {"id": "003000000000001", "success": True},
                {"id": "003000000000002", "success": True},
            ],
            status=200,
        )
        responses.add(
            responses.POST,
            url="https://example.com/services/data/v48.0/composite/sobjects",
            json=[{"id": "003000000000003", "success": True}],
            status=200,
        )

        recs = [
            ["Narvaez", "true"],
            ["Chalmers", "True"],
            ["De Vries", "False"],
            ["Aito", "false"],
            ["Boone", None],
        ]

        dml_op = RestApiDmlOperation(
            sobject="Contact",
            operation=DataOperationType.INSERT,
            context=task,
            api_options={},
            fields=["LastName", "IsEmailBounced"],
        )

        dml_op.start()
        dml_op.load_records(iter(recs))
        dml_op.end()

        assert json.loads(responses.calls[1].request.body) == {
            "allOrNone": False,
            "records": [
                {
                    "LastName": "Narvaez",
                    "IsEmailBounced": True,
                    "attributes": {"type": "Contact"},
                },
                {
                    "LastName": "Chalmers",
                    "IsEmailBounced": True,
                    "attributes": {"type": "Contact"},
                },
                {
                    "LastName": "De Vries",
                    "IsEmailBounced": False,
                    "attributes": {"type": "Contact"},
                },
                {
                    "LastName": "Aito",
                    "IsEmailBounced": False,
                    "attributes": {"type": "Contact"},
                },
                {
                    "LastName": "Boone",
                    "IsEmailBounced": False,
                    "attributes": {"type": "Contact"},
                },
            ],
        }

    @responses.activate
    def test_insert_dml_operation__row_failure(self):
        mock_describe_calls()
        task = _make_task(
            LoadData,
            {
                "options": {
                    "database_url": "sqlite:///test.db",
                    "mapping": "mapping.yml",
                }
            },
        )
        task.project_config.project__package__api_version = "48.0"
        task._init_task()

        responses.add(
            responses.POST,
            url="https://example.com/services/data/v48.0/composite/sobjects",
            json=[
                {"id": "003000000000001", "success": True},
                {"id": "003000000000002", "success": True},
            ],
            status=200,
        )
        responses.add(
            responses.POST,
            url="https://example.com/services/data/v48.0/composite/sobjects",
            json=[
                {
                    "id": "003000000000003",
                    "success": False,
                    "errors": [
                        {
                            "statusCode": "VALIDATION_ERR",
                            "message": "Bad data",
                            "fields": ["FirstName"],
                        }
                    ],
                }
            ],
            status=200,
        )

        recs = [["Fred", "Narvaez"], [None, "De Vries"], ["Hiroko", "Aito"]]

        dml_op = RestApiDmlOperation(
            sobject="Contact",
            operation=DataOperationType.INSERT,
            api_options={"batch_size": 2},
            context=task,
            fields=["FirstName", "LastName"],
        )

        dml_op.start()
        dml_op.load_records(iter(recs))
        dml_op.end()

        assert dml_op.job_result == DataOperationJobResult(
            DataOperationStatus.ROW_FAILURE, [], 3, 1
        )
        assert list(dml_op.get_results()) == [
            DataOperationResult("003000000000001", True, ""),
            DataOperationResult("003000000000002", True, ""),
            DataOperationResult(
                "003000000000003", False, "VALIDATION_ERR: Bad data (FirstName)"
            ),
        ]

    @responses.activate
    def test_insert_dml_operation__delete(self):
        mock_describe_calls()
        task = _make_task(
            LoadData,
            {
                "options": {
                    "database_url": "sqlite:///test.db",
                    "mapping": "mapping.yml",
                }
            },
        )
        task.project_config.project__package__api_version = "48.0"
        task._init_task()

        responses.add(
            responses.DELETE,
            url="https://example.com/services/data/v48.0/composite/sobjects?ids=003000000000001,003000000000002",
            json=[
                {"id": "003000000000001", "success": True},
                {"id": "003000000000002", "success": True},
            ],
            status=200,
        )
        responses.add(
            responses.DELETE,
            url="https://example.com/services/data/v48.0/composite/sobjects?ids=003000000000003",
            json=[{"id": "003000000000003", "success": True}],
            status=200,
        )

        recs = [["003000000000001"], ["003000000000002"], ["003000000000003"]]

        dml_op = RestApiDmlOperation(
            sobject="Contact",
            operation=DataOperationType.DELETE,
            api_options={"batch_size": 2},
            context=task,
            fields=["Id"],
        )

        dml_op.start()
        dml_op.load_records(iter(recs))
        dml_op.end()

        assert dml_op.job_result == DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 3, 0
        )
        assert list(dml_op.get_results()) == [
            DataOperationResult("003000000000001", True, ""),
            DataOperationResult("003000000000002", True, ""),
            DataOperationResult("003000000000003", True, ""),
        ]

    @responses.activate
    def test_insert_dml_operation__booleans(self):
        mock_describe_calls()
        task = _make_task(
            LoadData,
            {
                "options": {
                    "database_url": "sqlite:///test.db",
                    "mapping": "mapping.yml",
                }
            },
        )
        task.project_config.project__package__api_version = "48.0"
        task._init_task()

        responses.add(
            responses.POST,
            url="https://example.com/services/data/v48.0/composite/sobjects",
            json=[{"id": "003000000000001", "success": True}],
            status=200,
        )

        recs = [["Narvaez", "True"]]
        dml_op = RestApiDmlOperation(
            sobject="Contact",
            operation=DataOperationType.INSERT,
            api_options={"batch_size": 2},
            context=task,
            fields=["LastName", "IsEmailBounced"],  # IsEmailBounced is a Boolean field.
        )

        dml_op.start()
        dml_op.load_records(iter(recs))
        dml_op.end()

        json_body = json.loads(responses.calls[1].request.body)
        assert json_body["records"] == [
            {
                "LastName": "Narvaez",
                "IsEmailBounced": True,
                "attributes": {"type": "Contact"},
            }
        ]


class TestGetOperationFunctions:
    @mock.patch("cumulusci.tasks.bulkdata.step.BulkApiQueryOperation")
    @mock.patch("cumulusci.tasks.bulkdata.step.RestApiQueryOperation")
    def test_get_query_operation(self, rest_query, bulk_query):
        context = mock.Mock()
        context.sf.sf_version = "42.0"
        op = get_query_operation(
            sobject="Test",
            fields=["Id"],
            api_options={},
            context=context,
            query="SELECT Id FROM Test",
            api=DataApi.BULK,
        )
        assert op == bulk_query.return_value
        bulk_query.assert_called_once_with(
            sobject="Test",
            api_options={},
            context=context,
            query="SELECT Id FROM Test",
        )

        op = get_query_operation(
            sobject="Test",
            fields=["Id"],
            api_options={},
            context=context,
            query="SELECT Id FROM Test",
            api=DataApi.REST,
        )
        assert op == rest_query.return_value
        rest_query.assert_called_once_with(
            sobject="Test",
            fields=["Id"],
            api_options={},
            context=context,
            query="SELECT Id FROM Test",
        )

    @mock.patch("cumulusci.tasks.bulkdata.step.BulkApiQueryOperation")
    @mock.patch("cumulusci.tasks.bulkdata.step.RestApiQueryOperation")
    def test_get_query_operation__smart_to_rest(self, rest_query, bulk_query):
        context = mock.Mock()
        context.sf.restful.return_value = {"sObjects": [{"name": "Test", "count": 1}]}
        context.sf.sf_version = "42.0"
        op = get_query_operation(
            sobject="Test",
            fields=["Id"],
            api_options={},
            context=context,
            query="SELECT Id FROM Test",
            api=DataApi.SMART,
        )
        assert op == rest_query.return_value

        bulk_query.assert_not_called()
        context.sf.restful.assert_called_once_with("limits/recordCount?sObjects=Test")

    @mock.patch("cumulusci.tasks.bulkdata.step.BulkApiQueryOperation")
    @mock.patch("cumulusci.tasks.bulkdata.step.RestApiQueryOperation")
    def test_get_query_operation__smart_to_bulk(self, rest_query, bulk_query):
        context = mock.Mock()
        context.sf.restful.return_value = {
            "sObjects": [{"name": "Test", "count": 10000}]
        }
        context.sf.sf_version = "42.0"
        op = get_query_operation(
            sobject="Test",
            fields=["Id"],
            api_options={},
            context=context,
            query="SELECT Id FROM Test",
            api=DataApi.SMART,
        )
        assert op == bulk_query.return_value

        rest_query.assert_not_called()
        context.sf.restful.assert_called_once_with("limits/recordCount?sObjects=Test")

    @mock.patch("cumulusci.tasks.bulkdata.step.BulkApiQueryOperation")
    @mock.patch("cumulusci.tasks.bulkdata.step.RestApiQueryOperation")
    def test_get_query_operation__old_api(self, rest_query, bulk_query):
        context = mock.Mock()
        context.sf.sf_version = "39.0"
        op = get_query_operation(
            sobject="Test",
            fields=["Id"],
            api_options={},
            context=context,
            query="SELECT Id FROM Test",
            api=DataApi.SMART,
        )
        assert op == bulk_query.return_value

        context.sf.restful.assert_not_called()

    @mock.patch("cumulusci.tasks.bulkdata.step.BulkApiDmlOperation")
    @mock.patch("cumulusci.tasks.bulkdata.step.RestApiDmlOperation")
    def test_get_dml_operation(self, rest_dml, bulk_dml):
        context = mock.Mock()
        context.sf.sf_version = "42.0"
        op = get_dml_operation(
            sobject="Test",
            operation=DataOperationType.INSERT,
            fields=["Name"],
            api_options={},
            context=context,
            api=DataApi.BULK,
            volume=1,
        )

        assert op == bulk_dml.return_value
        bulk_dml.assert_called_once_with(
            sobject="Test",
            operation=DataOperationType.INSERT,
            fields=["Name"],
            api_options={},
            context=context,
        )

        op = get_dml_operation(
            sobject="Test",
            operation=DataOperationType.INSERT,
            fields=["Name"],
            api_options={},
            context=context,
            api=DataApi.REST,
            volume=1,
        )

        assert op == rest_dml.return_value
        rest_dml.assert_called_once_with(
            sobject="Test",
            operation=DataOperationType.INSERT,
            fields=["Name"],
            api_options={},
            context=context,
        )

    @mock.patch("cumulusci.tasks.bulkdata.step.BulkApiDmlOperation")
    @mock.patch("cumulusci.tasks.bulkdata.step.RestApiDmlOperation")
    def test_get_dml_operation__smart(self, rest_dml, bulk_dml):
        context = mock.Mock()
        context.sf.sf_version = "42.0"
        assert (
            get_dml_operation(
                sobject="Test",
                operation=DataOperationType.INSERT,
                fields=["Name"],
                api_options={},
                context=context,
                api=DataApi.SMART,
                volume=1,
            )
            == rest_dml.return_value
        )

        assert (
            get_dml_operation(
                sobject="Test",
                operation=DataOperationType.INSERT,
                fields=["Name"],
                api_options={},
                context=context,
                api=DataApi.SMART,
                volume=10000,
            )
            == bulk_dml.return_value
        )

        assert (
            get_dml_operation(
                sobject="Test",
                operation=DataOperationType.HARD_DELETE,
                fields=["Name"],
                api_options={},
                context=context,
                api=DataApi.SMART,
                volume=1,
            )
            == bulk_dml.return_value
        )

    @mock.patch("cumulusci.tasks.bulkdata.step.BulkApiDmlOperation")
    @mock.patch("cumulusci.tasks.bulkdata.step.RestApiDmlOperation")
    def test_get_dml_operation__old_api(self, rest_dml, bulk_dml):
        context = mock.Mock()
        context.sf.sf_version = "39.0"
        assert (
            get_dml_operation(
                sobject="Test",
                operation=DataOperationType.INSERT,
                fields=["Name"],
                api_options={},
                context=context,
                api=DataApi.SMART,
                volume=1,
            )
            == bulk_dml.return_value
        )



from unittest.mock import Mock
import base64
import io
import json
import os
import zipfile

from cumulusci.tasks.salesforce.org_settings import DeployOrgSettings
from cumulusci.utils import temporary_dir
from .util import create_task


class TestDeployOrgSettings:
    def test_run_task(self):
        with temporary_dir() as d:
            with open("dev.json", "w") as f:
                json.dump(
                    {
                        "settings": {
                            "orgPreferenceSettings": {"s1DesktopEnabled": True},
                            "communitiesSettings": {"enableNetworksEnabled": True},
                        }
                    },
                    f,
                )
            path = os.path.join(d, "dev.json")
            task_options = {"definition_file": path, "api_version": "48.0"}
            task = create_task(DeployOrgSettings, task_options)
            task.api_class = Mock()
            task()

        package_zip = task.api_class.call_args[0][1]
        zf = zipfile.ZipFile(io.BytesIO(base64.b64decode(package_zip)), "r")
        assert (
            readtext(zf, "package.xml")
            == """<?xml version="1.0" encoding="UTF-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <types>
        <members>*</members>
        <name>Settings</name>
    </types>
    <version>48.0</version>
</Package>"""
        )
        assert (
            readtext(zf, "settings/OrgPreference.settings")
            == """<?xml version="1.0" encoding="UTF-8"?>
<OrgPreferenceSettings xmlns="http://soap.sforce.com/2006/04/metadata">
    <preferences>
        <settingName>S1DesktopEnabled</settingName>
        <settingValue>True</settingValue>
    </preferences>
</OrgPreferenceSettings>"""
        )
        assert (
            readtext(zf, "settings/Communities.settings")
            == """<?xml version="1.0" encoding="UTF-8"?>
<CommunitiesSettings xmlns="http://soap.sforce.com/2006/04/metadata">
    <enableNetworksEnabled>True</enableNetworksEnabled>
</CommunitiesSettings>"""
        )

    def test_run_task__no_settings(self):
        with temporary_dir() as d:
            with open("dev.json", "w") as f:
                f.write("{}")
            path = os.path.join(d, "dev.json")
            task_options = {"definition_file": path}
            task = create_task(DeployOrgSettings, task_options)
            task.api_class = Mock()
            task()
            assert not task.api_class.called


def readtext(zf, name):
    with zf.open(name, "r") as f:
        return io.TextIOWrapper(f).read()



from typing import Optional

from cumulusci.tasks.metadata_etl import MetadataSingleEntityTransformTask
from cumulusci.utils.xml.metadata_tree import MetadataElement
from cumulusci.core.utils import process_bool_arg


class SetDuplicateRuleStatus(MetadataSingleEntityTransformTask):
    entity = "DuplicateRule"
    task_options = {
        "active": {
            "description": "Boolean value, set the Duplicate Rule to either active or inactive",
            "required": True,
        },
        **MetadataSingleEntityTransformTask.task_options,
    }

    def _transform_entity(
        self, metadata: MetadataElement, api_name: str
    ) -> Optional[MetadataElement]:
        status = "true" if process_bool_arg(self.options["active"]) else "false"
        metadata.find("isActive").text = status

        return metadata



from cumulusci.core.exceptions import TaskOptionsError
from unittest.mock import Mock
import unittest

from cumulusci.tasks.preflight.sobjects import CheckSObjectOWDs, CheckSObjectsAvailable
from cumulusci.tasks.salesforce.tests.util import create_task

from simple_salesforce.exceptions import SalesforceMalformedRequest


class TestCheckSObjectsAvailable(unittest.TestCase):
    def test_sobject_preflight(self):
        task = create_task(CheckSObjectsAvailable, {})

        task._init_task = Mock()
        task.sf = Mock()
        task.sf.describe.return_value = {
            "sobjects": [{"name": "Network"}, {"name": "Account"}]
        }

        task()

        assert task.return_values == {"Network", "Account"}


class TestCheckSObjectOWDs(unittest.TestCase):
    def test_sobject_preflight__positive(self):
        task = create_task(
            CheckSObjectOWDs,
            {
                "org_wide_defaults": [
                    {"api_name": "Account", "internal_sharing_model": "Private"},
                    {"api_name": "Contact", "internal_sharing_model": "ReadWrite"},
                ]
            },
        )
        task._init_task = Mock()
        task.sf = Mock()
        task.sf.query.return_value = {
            "totalSize": 2,
            "records": [
                {"QualifiedApiName": "Account", "InternalSharingModel": "Private"},
                {"QualifiedApiName": "Contact", "InternalSharingModel": "ReadWrite"},
            ],
        }
        task()

        assert task.return_values is True

    def test_sobject_preflight__negative(self):
        task = create_task(
            CheckSObjectOWDs,
            {
                "org_wide_defaults": [
                    {"api_name": "Account", "internal_sharing_model": "Private"}
                ]
            },
        )
        task._init_task = Mock()
        task.sf = Mock()
        task.sf.query.return_value = {
            "totalSize": 1,
            "records": [
                {"QualifiedApiName": "Account", "InternalSharingModel": "ReadWrite"}
            ],
        }
        task()

        assert task.return_values is False

    def test_sobject_preflight__external(self):
        task = create_task(
            CheckSObjectOWDs,
            {
                "org_wide_defaults": [
                    {"api_name": "Account", "external_sharing_model": "Private"}
                ]
            },
        )
        task._init_task = Mock()
        task.sf = Mock()
        task.sf.query.return_value = {
            "totalSize": 1,
            "records": [
                {
                    "QualifiedApiName": "Account",
                    "InternalSharingModel": "Private",
                    "ExternalSharingModel": "Private",
                }
            ],
        }
        task()

        assert task.return_values is True

    def test_sobject_preflight__both(self):
        task = create_task(
            CheckSObjectOWDs,
            {
                "org_wide_defaults": [
                    {
                        "api_name": "Account",
                        "internal_sharing_model": "ReadWrite",
                        "external_sharing_model": "Private",
                    }
                ]
            },
        )
        task._init_task = Mock()
        task.sf = Mock()
        task.sf.query.return_value = {
            "totalSize": 1,
            "records": [
                {
                    "QualifiedApiName": "Account",
                    "InternalSharingModel": "ReadWrite",
                    "ExternalSharingModel": "Private",
                }
            ],
        }
        task()

        assert task.return_values is True

    def test_sobject_preflight__external_not_enabled(self):
        task = create_task(
            CheckSObjectOWDs,
            {
                "org_wide_defaults": [
                    {"api_name": "Account", "external_sharing_model": "Private"}
                ]
            },
        )
        task._init_task = Mock()
        task.sf = Mock()
        task.sf.query.side_effect = SalesforceMalformedRequest(
            "url", 400, "resource_name", "content"
        )
        task()

        assert task.return_values is False

    def test_sobject_preflight__task_options(self):
        with self.assertRaises(TaskOptionsError):
            create_task(CheckSObjectOWDs, {})
        with self.assertRaises(TaskOptionsError):
            create_task(
                CheckSObjectOWDs,
                {"org_wide_defaults": [{"internal_sharing_model": "Private"}]},
            )
        with self.assertRaises(TaskOptionsError):
            create_task(
                CheckSObjectOWDs, {"org_wide_defaults": [{"api_name": "Account"}]}
            )



from cumulusci.tasks.salesforce import BaseSalesforceApiTask
from cumulusci.core.exceptions import TaskOptionsError

from simple_salesforce.exceptions import SalesforceMalformedRequest


class CheckSObjectsAvailable(BaseSalesforceApiTask):
    api_version = "48.0"

    def _run_task(self):
        self.return_values = {entry["name"] for entry in self.sf.describe()["sobjects"]}

        self.logger.info(
            "Completed sObjects preflight check with result {}".format(
                self.return_values
            )
        )


class CheckSObjectOWDs(BaseSalesforceApiTask):
    task_options = {
        "org_wide_defaults": {
            "description": "The Organization-Wide Defaults to check, "
            "organized as a list with each element containing the keys api_name, "
            "internal_sharing_model, and external_sharing_model. NOTE: you must have "
            "External Sharing Model turned on in Sharing Settings to use the latter feature. "
            "Checking External Sharing Model when it is turned off will fail the preflight.",
            "required": True,
        }
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)

        if "org_wide_defaults" not in self.options:
            raise TaskOptionsError("org_wide_defaults is a required option")

        if not all("api_name" in entry for entry in self.options["org_wide_defaults"]):
            raise TaskOptionsError("api_name must be included in each entry")

        if not all(
            "internal_sharing_model" in entry or "external_sharing_model" in entry
            for entry in self.options["org_wide_defaults"]
        ):
            raise TaskOptionsError("Each entry must include a sharing model to check.")

        self.owds = {
            entry["api_name"]: (
                entry.get("internal_sharing_model"),
                entry.get("external_sharing_model"),
            )
            for entry in self.options["org_wide_defaults"]
        }

    def _check_owds(self, sobject, result):
        internal = (
            result["InternalSharingModel"] == self.owds[sobject][0]
            if self.owds[sobject][0]
            else True
        )
        external = (
            result["ExternalSharingModel"] == self.owds[sobject][1]
            if self.owds[sobject][1]
            else True
        )
        return internal and external

    def _run_task(self):
        try:
            ext = (
                ", ExternalSharingModel"
                if any(owd[1] is not None for owd in self.owds.values())
                else ""
            )
            object_list = ", ".join(f"'{obj}'" for obj in self.owds.keys())
            results = self.sf.query(
                f"SELECT QualifiedApiName, InternalSharingModel{ext} "
                "FROM EntityDefinition "
                f"WHERE QualifiedApiName IN ({object_list})"
            )["records"]
            self.return_values = all(
                self._check_owds(rec["QualifiedApiName"], rec) for rec in results
            )
        except (IndexError, KeyError, SalesforceMalformedRequest):
            self.return_values = False

        self.logger.info(
            f"Completed Organization-Wide Default preflight with result: {self.return_values}"
        )



import doctest
from io import BytesIO, UnsupportedOperation
from pathlib import Path
from unittest import mock
from tempfile import TemporaryDirectory
import os
import sys

import pytest
import responses
from fs import open_fs, errors

import cumulusci
from cumulusci.utils import fileutils, temporary_dir
from cumulusci.utils.fileutils import (
    load_from_source,
    open_fs_resource,
    FSResource,
)


class TestFileutils:
    @responses.activate
    @mock.patch("urllib.request.urlopen")
    def test_docstrings(self, urlopen):
        pretend_html = b"<!DOCTYPE HTML ...blah blah blah"
        responses.add("GET", "http://www.salesforce.com/", body=pretend_html)
        fake_http_stream = BytesIO(pretend_html)
        fake_http_stream.url = "https://www.salesforce.com/"
        urlopen.return_value = fake_http_stream
        try:
            doctest.testmod(fileutils, raise_on_error=True, verbose=True)
        except doctest.DocTestFailure as e:
            print("Got")
            print(str(e.got))
            raise

    def test_binary_becomes_string(self):
        with load_from_source(BytesIO(b"foo")) as (path, file):
            data = file.read()
            assert isinstance(data, str)
            assert data == "foo"

    def test_writable_file_throws(self):
        with temporary_dir():
            with open("writable", "wt") as t:
                with pytest.raises(UnsupportedOperation):
                    with load_from_source(t) as (filename, data):
                        pass

    @responses.activate
    def test_load_from_url(self):
        html = "<!DOCTYPE HTML ..."
        responses.add("GET", "http://www.salesforce.com", body=html)
        with load_from_source("http://www.salesforce.com") as (filename, data):
            assert data.read() == html

    def test_load_from_Path(self):
        p = Path(cumulusci.__file__).parent / "cumulusci.yml"
        with load_from_source(p) as (filename, data):
            assert "tasks:" in data.read()

    def test_load_from_path_string(self):
        p = Path(cumulusci.__file__).parent / "cumulusci.yml"
        with load_from_source(str(p)) as (filename, data):
            assert "tasks:" in data.read()

    def test_load_from_open_file(self):
        p = Path(cumulusci.__file__).parent / "cumulusci.yml"
        with open(p) as f:
            with load_from_source(f) as (filename, data):
                assert "tasks:" in data.read()
                assert str(p) == filename

    def test_load_from_fs_resource(self):
        p = Path(cumulusci.__file__).parent / "cumulusci.yml"
        with open_fs_resource(p) as p2:
            with load_from_source(p2) as (filename, data):
                assert "tasks:" in data.read()


class _TestFSResourceShared:
    file = Path(__file__)

    def test_resource_exists_abspath(self):
        abspath = os.path.abspath(self.file)
        with open_fs_resource(abspath) as resource:
            assert resource.exists()
            assert str(resource.getsyspath()) == abspath

    def test_resource_does_not_exist(self):
        abspath = os.path.abspath(self.file)
        with open_fs_resource(abspath + "xyzzy") as resource:
            assert not resource.exists()
            assert abspath + "xyzzy" in resource

    def test_resource_exists_pathlib_abspath(self):
        abspath = os.path.abspath(self.file)
        with open_fs_resource(Path(abspath)) as resource:
            assert resource.exists()
            assert abspath in resource

    def test_resource_doesnt_exist_pathlib_abspath(self):
        abspath = os.path.abspath(self.file)
        with open_fs_resource(Path(abspath + "xyzzy")) as resource:
            assert not resource.exists()
            assert abspath + "xyzzy" in resource

    def test_resource_exists_url_abspath(self):
        abspath = os.path.abspath(self.file)
        url = f"file://{abspath}"
        with open_fs_resource(url) as resource:
            assert resource.exists()
            assert abspath in resource

    def test_resource_as_str(self):
        abspath = os.path.abspath(self.file)
        with open_fs_resource(Path(abspath)) as resource:
            assert "file://" in repr(resource)

    def test_join_paths(self):
        parent = Path(self.file).parent
        with open_fs_resource(parent) as resource:
            this_file = resource / self.file.name
            assert this_file.exists()

    def test_clone_fsresource(self):
        abspath = os.path.abspath(self.file)
        with open_fs_resource(Path(abspath)) as resource:
            with open_fs_resource(resource) as resource2:
                assert abspath in str(resource2)

    def test_load_from_file_system(self):
        abspath = os.path.abspath(self.file)
        fs = open_fs("/")
        with open_fs_resource(abspath, fs) as f:
            assert abspath in str(f)

    def test_windows_path(self):
        abspath = "c:\\foo\\bar"
        with open_fs_resource(abspath) as f:
            if sys.platform == "win32":
                assert "c:\\foo\\bar" == str(f.getsyspath()), str(f.getsyspath())


class TestFSResource(_TestFSResourceShared):
    def test_resource_exists_relpath(self):
        relpath = os.path.relpath(self.file)
        with open_fs_resource(relpath) as resource:
            assert resource.exists()
            assert (
                str(Path(resource.getsyspath()).relative_to(Path(".").absolute()))
                == relpath
            )

    def test_resource_doesnt_exist_relpath(self):
        relpath = os.path.relpath(self.file)
        with open_fs_resource(relpath + "xyzzy") as resource:
            assert not resource.exists()
            assert relpath + "xyzzy" in resource

    def test_resource_exists_pathlib_relpath(self):
        relpath = os.path.relpath(self.file)
        with open_fs_resource(Path(relpath)) as resource:
            assert resource.exists()
            assert relpath in resource

    def test_resource_exists_url_relpath(self):
        relpath = os.path.relpath(self.file)
        url = f"file://{relpath}"
        with open_fs_resource(url) as resource:
            assert resource.exists()
            assert relpath in resource

    def test_resource_test_resource_doesnt_exist_pathlib_relpath(self):
        relpath = os.path.relpath(self.file)
        with open_fs_resource(Path(relpath + "xyzzy")) as resource:
            assert not resource.exists()
            assert relpath + "xyzzy" in resource


class TestFSResourceTempdir(_TestFSResourceShared):
    def setup(self):
        self.tempdir = TemporaryDirectory()
        self.file = Path(self.tempdir.name) / "testfile.txt"
        self.file.touch()

    def teardown(self):
        self.tempdir.cleanup()

    def test_copy_to(self):
        abspath = os.path.abspath(self.file)
        with open_fs_resource(abspath) as f:
            f.copy_to(abspath + ".bak")
            assert Path(abspath + ".bak").exists()
            f.copy_to(Path(abspath + ".bak2"))
            assert Path(abspath + ".bak2").exists()
            f.copy_to(FSResource.new(abspath + ".bak3"))
            assert Path(abspath + ".bak3").exists()

    def test_mkdir_rmdir(self):
        abspath = Path(self.tempdir.name) / "doesnotexist"
        assert not abspath.exists()
        with open_fs_resource(abspath) as f:
            f.mkdir()
            assert abspath.exists()
            f.mkdir(exist_ok=True)
            f.rmdir()

        abspath = abspath / "foo" / "bar" / "baz"
        assert not abspath.exists()
        with open_fs_resource(abspath) as f:
            f.mkdir(parents=True)
            f.mkdir(parents=True, exist_ok=True)
            f.mkdir(parents=False, exist_ok=True)
            assert abspath.exists()

            with pytest.raises(errors.DirectoryExists):
                f.mkdir(parents=False, exist_ok=False)
            f.rmdir()

    def test_open_for_write(self):
        abspath = Path(self.tempdir.name) / "blah"
        with open_fs_resource(abspath) as fs:
            fs.mkdir()
            newfile = fs / "newfile"
            with newfile.open("w") as f:
                f.write("xyzzy")

            with newfile.open("r") as f:
                assert f.read() == "xyzzy"


class TestFSResourceError:
    def test_fs_resource_init_error(self):
        with pytest.raises(NotImplementedError):
            FSResource()
